/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <stdio.h> 
#include <stddef.h> 
#include <stdlib.h> 
#include "arith.h"    /* Get necessary limits */
#include "prob_dist.h"
#include "epwic.h"
#include "pyramid_util.h"

#define NR_END 1 
#define FREE_ARG char* 
#define JTRY(j,k,m) ((long)((((double)(k))*((double)(j)))/((double)(m))))
#define IMAX(i,j) ((i>j)?i:j)


void nrerror(char error_text[]) 
/* Numerical Recipes standard error handler */ 
{
  fprintf(stderr,"Numerical Recipes run-time error...\n"); 
  fprintf(stderr,"%s\n",error_text); 
  fprintf(stderr,"...now exiting to system...\n"); 
  exit(1); 
}


unsigned long *lvector(long nl, long nh) 
/* allocate an unsigned long vector with subscript range v[nl..nh] */ 
{ 
  unsigned long *v; 
  v=(unsigned long *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(long))); 
  if (!v) nrerror("allocation failure in lvector()"); 
  return v-nl+NR_END; 
}


void free_lvector(unsigned long *v, long nl, long nh) 
/* free an unsigned long vector allocated with lvector() */ 
{ 
  free((FREE_ARG) (v+nl-NR_END)); 
}


/* Given a table nfreq[1..nchh] of the frequency of occurrence of nchh
   symbols, and given a desired output radix nradd, initialize the
   cumulative frequency table and other variables for
   arithmetic compression in the structure acode. */

void arcmak(unsigned long nradd, arithcode *acode) 
{ 
  int j;
  
  if(nradd > 256) nrerror("output radix may not exceed 256 in arcmak.");

  acode->minint=MAXINT/nradd; 
  acode->nrad=nradd;
  acode->jdif=acode->nrad-1; 
  acode->lcd = 0;

  for (j=NWK;j>=1;j--) { 
    acode->iupb[j]=acode->nrad-1; 
    acode->ilob[j]=0; 
    acode->nc=j; 
    if (acode->jdif > acode->minint) return; /* Initialization complete. */
    acode->jdif=(acode->jdif+1)*acode->nrad-1; 
  } 
  nrerror("NWK too small in arcode.");
}


/* Used by arith_encode and arith_decode. Add the integer ja to the
   radix nrad multiple-precision integer iin[nc..nwk]. Return the
   result in iout[nc..nwk]. */

void arcsum(unsigned long iin[], unsigned long iout[], unsigned long ja, 
	    int nwk, unsigned long nrad, unsigned long nc)
{ 
  int j,karry=0; 
  unsigned long jtmp; 
  for (j=nwk;j>nc;j--) { 
    jtmp=ja; 
    ja /= nrad; 
    iout[j]=iin[j]+(jtmp-ja*nrad)+karry; 
    if (iout[j] >= nrad) { 
      iout[j] -= nrad; 
      karry=1; 
    } else karry=0; 
  } 
  iout[nc]=iin[nc]+ja+karry; 
}

/* Compress (isign = 1) or decompress (isign = -1) the single
   character ich into or out of the character array *codep[1..lcode],
   starting with byte *codep[lcd] and (if necessary) incrementing lcd
   so that, on return, lcd points to the rst unused byte in
   *codep. Note that the structure acode contains both information on
   the code, and also state information on the particular output being
   written into the array *codep.  An initializing call with isign=0
   is required before beginning any *codep array, whether for encoding
   or decoding. This is in addition to the initializing call to arcmak
   that is required to initialize the code itself.  A call with
   ich=nch (as set in arcmak) has the reserved meaning "end of
   message."
     We assume here that we are encoding binary data, and at every
   step we provide the probability that we get a 0 (represented by
   an integer between 0 and 1000) */

void arith_encode(unsigned long ich, int prob_zero, int last, 
		  arithcode *acode)
{ 
  int j,k; 
  unsigned long jh,jl;

  if (ich >= 2) nrerror("bad ich in arcode."); 

  /* Following code is common for encoding and decoding. Convert
     character ich to a new subrange [ilob,iupb). */
  if (ich == 0){
    jh=JTRY(acode->jdif,prob_zero,1000); 
    jl=JTRY(acode->jdif,0,1000); 
  } else {
    jh=JTRY(acode->jdif,1000,1000); 
    jl=JTRY(acode->jdif,prob_zero,1000);
  }
  acode->jdif=jh-jl; 
  arcsum(acode->ilob,acode->iupb,jh,NWK,acode->nrad,acode->nc); 
  arcsum(acode->ilob,acode->ilob,jl,NWK,acode->nrad,acode->nc); 

  /* How many leading digits to output (if encoding) or skip over? */
  for (j=acode->nc;j<=NWK;j++) {
    if (!last && acode->iupb[j] != acode->ilob[j]) break; 
    if (acode->lcd > acode->lcode) { 
      fprintf(stderr,"Reached the end of the 'code' array.\n"); 
      fprintf(stderr,"Attempting to expand its size.\n"); 
      acode->lcode *= 1.5; 
      if ((acode->codep=(unsigned char *)realloc(acode->codep, 
          (unsigned)((acode->lcode)*sizeof(unsigned char)))) == NULL) { 
	nrerror("Size expansion failed"); 
      } 
    } 
    (acode->codep)[acode->lcd]=(unsigned char)acode->ilob[j]; 
    ++(acode->lcd); 
  } 
  if (j > NWK) return; 
  /* Ran out of message. Did someone forget to use "last"? */
  acode->nc=j; 
  for(j=0;acode->jdif<acode->minint;j++) /* How many digits to shift? */
    acode->jdif *= acode->nrad; 
  if (acode->nc-j < 1) nrerror("NWK too small in arcode."); 
  if (j) { /* Shift them. */
    for (k=acode->nc;k<=NWK;k++) { 
      acode->iupb[k-j]=acode->iupb[k]; 
      acode->ilob[k-j]=acode->ilob[k]; 
    } 
  } 
  acode->nc -= j; 
  for (k=NWK-j+1;k<=NWK;k++) acode->iupb[k]=acode->ilob[k]=0; 
  return; /* Normal return. */
}

void arith_decode(unsigned long *ich, int prob_zero, arithcode *acode)
{ 
  int j,k; 
  unsigned long ja,jh,jl;

  ja=(acode->codep)[acode->lcd] - acode->ilob[acode->nc]; 
  for (j=acode->nc+1;j<=NWK;j++) { 
    ja *= acode->nrad; 
    ja += ((acode->codep)[acode->lcd+j-acode->nc]-acode->ilob[j]); 
  } 
  if(ja >= JTRY(acode->jdif,prob_zero,1000)){
    *ich=1;
  } else {
    *ich=0;
  }

  /* Following code is common for encoding and decoding. Convert
     character ich to a new subrange [ilob,iupb). */
  if (*ich == 0){
    jh=JTRY(acode->jdif,prob_zero,1000); 
    jl=JTRY(acode->jdif,0,1000); 
  } else {
    jh=JTRY(acode->jdif,1000,1000); 
    jl=JTRY(acode->jdif,prob_zero,1000);
  }
  acode->jdif=jh-jl; 
  arcsum(acode->ilob,acode->iupb,jh,NWK,acode->nrad,acode->nc); 
  arcsum(acode->ilob,acode->ilob,jl,NWK,acode->nrad,acode->nc); 
  
  /* How many leading digits to skip over? */
  for (j=acode->nc;j<=NWK;j++) {
    if (acode->iupb[j] != acode->ilob[j]) break; 
    ++(acode->lcd); 
  } 
  if (j > NWK) return; 
  /* Ran out of message. Did someone forget to use "last"? */
  acode->nc=j; 
  for(j=0;acode->jdif<acode->minint;j++) /* How many digits to shift? */
    acode->jdif *= acode->nrad; 
  if (acode->nc-j < 1) nrerror("NWK too small in arcode."); 
  if (j) { /* Shift them. */
    for (k=acode->nc;k<=NWK;k++) { 
      acode->iupb[k-j]=acode->iupb[k]; 
      acode->ilob[k-j]=acode->ilob[k]; 
    } 
  } 
  acode->nc -= j; 
  for (k=NWK-j+1;k<=NWK;k++) acode->iupb[k]=acode->ilob[k]=0; 
  return; /* Normal return. */
}

unsigned char adf_get_byte(FILE *infile, int loc, arithcode *acode, 
			   int *bytes_read)
{
  while(acode->lcd <= loc){
    if(fread(&((acode->codep)[acode->lcd]),1,1,infile) <= 0){
      nrerror("File read error");
    }
    (*bytes_read)++;
    (acode->lcd)++;
  }
  return ((acode->codep)[loc]);
}

/* Here we read from infile exactly those bytes necessary to
   decode it.  Note that by reading bytes, it is best if the
   radix of the coding is 256 to completely utilize space.
   Here, lcd is used to note the location of the last read
   byte in codep.  Finally, codep is used to temporarily store
   the bytes from infile and should be at least NWK size large.
   It returns the number of characters read */

int arith_decode_file (FILE *infile, unsigned long *ich, 
    int prob_zero, arithcode *acode) 
{ 
  int j,k,l;
  unsigned long ja,jh,jl;
  int bytes_read = 0;

  ja=adf_get_byte(infile,0,acode,&bytes_read)-acode->ilob[acode->nc]; 
  for (j=acode->nc+1;j<=NWK;j++) { 
    ja *= acode->nrad; 
    ja += (adf_get_byte(infile,j-acode->nc,acode,&bytes_read)
	   -acode->ilob[j]); 
  } 
  if(ja >= JTRY(acode->jdif,prob_zero,1000)){
    *ich=1;
  } else {
    *ich=0;
  }

  /* Following code is common for encoding and decoding. Convert
     character ich to a new subrange [ilob,iupb). */
  if (*ich == 0){
    jh=JTRY(acode->jdif,prob_zero,1000); 
    jl=JTRY(acode->jdif,0,1000); 
  } else {
    jh=JTRY(acode->jdif,1000,1000); 
    jl=JTRY(acode->jdif,prob_zero,1000);
  }
  acode->jdif=jh-jl; 
  arcsum(acode->ilob,acode->iupb,jh,NWK,acode->nrad,acode->nc); 
  arcsum(acode->ilob,acode->ilob,jl,NWK,acode->nrad,acode->nc); 
  
  k=0;

  /* How many leading digits to skip over? */
  for (j=acode->nc;j<=NWK;j++) {
    if (acode->iupb[j] != acode->ilob[j]) break; 
    k++;
  }
 
  if (j > NWK) return(bytes_read); 

  /* Actually shift values */
  if (k>0){
    for(l=0;l<acode->lcd-k;l++){
      (acode->codep)[l] = (acode->codep)[l+k];
    }
    (acode->lcd)-=k;
  }

  acode->nc=j; 
  for(j=0;acode->jdif<acode->minint;j++) /* How many digits to shift? */
    acode->jdif *= acode->nrad; 
  if (acode->nc-j < 1) nrerror("NWK too small in arcode."); 
  if (j) { /* Shift them. */
    for (k=acode->nc;k<=NWK;k++) { 
      acode->iupb[k-j]=acode->iupb[k]; 
      acode->ilob[k-j]=acode->ilob[k]; 
    } 
  } 
  acode->nc -= j; 
  for (k=NWK-j+1;k<=NWK;k++) acode->iupb[k]=acode->ilob[k]=0; 
  return (bytes_read); /* Normal return. */
}

/* This routine takes a byte and encodes it with the probability
   1/2 of getting a 0 or 1 for each bit (therefore just encoding it
   straight).  The argument last states if the last bit should be
   considered to be the last bit in the encoded stream */
int arith_encode_byte(unsigned char ich, int last, arithcode *acode){
  int i;
  int bit;
  
  for (i=7;i>=0;i--){  /* output the byte left to right */
    bit = (int) ((ich >> i) & 1);
    arith_encode((unsigned long) bit, 500, ((i==0)&&last), acode);
  }
  return(acode->lcd);
}  

int arith_encode_block(unsigned char * block, int block_length,
			arithcode *acode){
  int i;
  for(i=0;i<block_length;i++){
    arith_encode_byte(block[i],0,acode);
  }
  return(acode->lcd);
}

int arith_decode_file_byte(FILE *infile, unsigned char * ich,
			    arithcode *acode){
  int i;
  unsigned long bit;
  unsigned char t_byte;
  int num_bytes;
  
  t_byte = 0;
  num_bytes = 0;
  for (i=7;i>=0;i--){  /* fill in byte left to right */
    num_bytes += arith_decode_file(infile,&bit,500,acode);
    t_byte <<= 1;
    t_byte |= (bit?1:0);
  }
  (*ich) = t_byte;
  return(num_bytes);
}


FILE *_arith_infile = NULL;
Pyramid _arith_pyr = NULL;
int _arith_i;
int _arith_subband;
int _arith_bytes_read;
int _arith_read_sign;
int _arith_mult;
arithcode * _arith_acode;
int *_arith_bit_level;

void arith_bitplane_init_decode(FILE *infile,Pyramid pyr, int subband,
				int *read_to_bit, arithcode *acode)
{
  unsigned char s,p;

  _arith_infile = infile;
  _arith_pyr = pyr;
  _arith_i = 0;
  _arith_subband = subband;
  _arith_bytes_read = 0;
  _arith_read_sign = 0;
  _arith_acode = acode;
  _arith_bit_level = read_to_bit;


  if (!exist_prior(_arith_subband)){
    _arith_bytes_read += 
      arith_decode_file_byte(_arith_infile,&s,_arith_acode);
    _arith_bytes_read +=
      arith_decode_file_byte(_arith_infile,&p,_arith_acode);
    calc_prior(_arith_subband,((double)s)*128.0/255.0 + 0.5,
	       ((double)p)*1.8/255.0 + 0.2);
  }
  _arith_mult = 128;
}

/* This extracts the abs value already read to the child */
void get_read_value(Coefficient value, int bit_level, int *abs){
  Coefficient mask;

  /* Kill sign bit */
  value &= COEF_ABS_MASK;

  if (bit_level >= 0){
    mask = (1<<(bit_level+1))-1;
    *abs = value & (~mask);
  } else {
    *abs = value;
  }
}  

int arith_bitplane_read_bit(){
  double parent_min,parent_max;
  int child_min,child_max;
  int child_bit_level;
  int prob,subband; 
  unsigned long read_bit;
  int child_x,child_y;
  Pyramid pyr;

  if (_arith_i >= pyrSubSize(_arith_pyr,_arith_subband)){
    return 0;
  }
  if (_arith_read_sign){
    _arith_i ++;
    _arith_read_sign = 0;
    _arith_bytes_read += arith_decode_file(_arith_infile,
			&read_bit, 500, _arith_acode);
    return ((int) read_bit);
  } else {
    subband = _arith_subband; /* Redifinitions for convenience */
    pyr = _arith_pyr;

    child_x = _arith_i % pyrSubXSize(_arith_pyr,subband);
    child_y = _arith_i / pyrSubXSize(_arith_pyr,subband);

    child_bit_level = _arith_bit_level[subband];
 
    parent_min = 0;
    parent_max = MAX_ABS_COEFF+1;

    get_read_value(pyrCoefCoord(_arith_pyr,subband,child_x,child_y),
                   child_bit_level,
		   &child_min);

    child_max = child_min + (1<<(child_bit_level+1));

    /* Calculate probability that the next bit is 0 */

    prob = (int) (1000 *
		  (get_prior_CDF(subband,child_min+(1<<child_bit_level))-
		   get_prior_CDF(subband,child_min))/
		  (get_prior_CDF(subband,child_max)-
		   get_prior_CDF(subband,child_min)));


    /* If the division results in a probability we can't believe,
       set the probability to 1/2 */

    if ((prob <=0) || (prob >1000)) {
      prob = 500;
    }

    if (prob == 1000){
      prob = 999;
    }

    _arith_bytes_read += arith_decode_file(_arith_infile,
			&read_bit, prob, _arith_acode);

    if ((child_min == 0) && (read_bit == 1)){
      _arith_read_sign = 1;
    } else {
      _arith_i ++;
    }
    return ((int) read_bit);
  }
}

void arith_bitplane_read_byte(Byte *sym){
  int i;
  int bit;
  unsigned char t_byte;
  
  t_byte = 0;
  for (i=7;i>=0;i--){  /* fill in byte left to right */
    bit = arith_bitplane_read_bit();
    t_byte <<= 1;
    t_byte |= (bit?1:0);
  }
  (*sym) = t_byte;
}

int arith_bitplane_bytes_read(){
  return (_arith_bytes_read);
}

