/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#ifndef _EPWIC_H
#define _EPWIC_H

#include <stdio.h>  /* for FILE */

/* 
For portability, we assume ONLY that the space allotted to a char is
at least 8 bits, that a short is at least two chars, and that an int
is at least 4 chars.  The compiler function sizeof() returns the size
of its arg in units of characters.
*/

typedef unsigned char Byte;

#define log2(x) (log(x)/log(2))

/* This is the constant of proportinality for which the salience depends
   on the level of the pyramid.  If the value is < 0, then we prompt
   the user for the value */
#define SALIENCE_LEVEL_MULT 0

/* Maximum x- or y-size of image (will be stored in a short) */
#define MAX_IMAGE_DIM 65535  

/* Log (base 2) of MAX_IMAGE_DIM^2 */
#define LOG_MAX_IMAGE_SIZE 32

/* This number must be consistent with the filters that are 
   hardwired into epwic.c */
#define FILTER_SIZE 9

/* Maximum number of pyramid levels (value 3*levs+1 stored in 5 bits) */
#define MAX_LEVELS 10

/* ============= FILE BYTE TAGS =============== */
#define EPWIC_ID_TAG   "W"
#define VERSION_TAG   '\020'        /* Version 1.0 (0001.0000) */

/* coded data block tags: */
#define DONE_DATA_TAG       '\000'
#define BIN_INFO_TAG        '\001'
#define SUBSAMPLE_INFO_TAG  '\002'

#define DATA_MASK           '\340'  /* 11100000 */

#define RAW_DATA_TAG        '\200'  /* 100xxxxx */
#define RLE_TYPE_ONE        '\240'  /* 101xxxxx 16 bits raw, 8 bits zero */
#define RLE_TYPE_TWO        '\300'  /* 110xxxxx 16 raw,     16 zero */
#define RLE_TYPE_THREE      '\340'  /* 111xxxxx  8 raw,     16 zero */

/* ============== FILE IO MACROS ============== */
/*  We write everything as bytes for portability. For the            */
/*  files to be machine independent, we write the bytes of           */
/*  longer quantities (like shorts and ints) in a fixed order.       */
/*  **** The arrays are written incorrectly and need to be fixed!    */
/*  Temporary variables are defined in utilities.c                   */

#define write_byte(val, stream) temp_byte = (Byte) val; \
  fwrite(&temp_byte,1,1,stream)
#define write_short(val,stream) temp_short = (short) val; \
  temp_byte = (temp_short & 0xFF00)>>8; fwrite(&temp_byte,1,1,stream); \
  temp_byte = (temp_short & 0x00FF); fwrite(&temp_byte,1,1,stream)
#define write_int(val,stream) temp_int = (unsigned int) val; \
  temp_byte = (temp_int & 0xFF000000)>>24; fwrite(&temp_byte,1,1,stream); \
  temp_byte = (temp_int & 0x00FF0000)>>16; fwrite(&temp_byte,1,1,stream); \
  temp_byte = (temp_int & 0x0000FF00)>>8;  fwrite(&temp_byte,1,1,stream); \
  temp_byte = (temp_int & 0x000000FF);     fwrite(&temp_byte,1,1,stream)


#define read_byte(sym, stream) fread(&temp_byte,1,1,stream); sym=temp_byte
#define read_short(sym, stream) \
  fread(&temp_byte,1,1,stream); temp_short = temp_byte; temp_short <<=8; \
  fread(&temp_byte,1,1,stream); temp_short |= temp_byte; sym=temp_short
#define read_int(sym, stream) \
  fread(&temp_byte,1,1,stream); temp_int  = temp_byte; temp_int <<= 8; \
  fread(&temp_byte,1,1,stream); temp_int |= temp_byte; temp_int <<= 8; \
  fread(&temp_byte,1,1,stream); temp_int |= temp_byte; temp_int <<= 8; \
  fread(&temp_byte,1,1,stream); temp_int |= temp_byte; sym = temp_int;


extern Byte temp_byte;
extern short temp_short;
extern int temp_int;

/* ============= FUNCTION DECLARATIONS ============= */
/*    These functions are defined in utilities.c     */
#include <stdio.h>
void write_array(void *ptr, int ptrsize, int arraysize, FILE* stream);
void read_array(void *ptr, int ptrsize, int arraysize, FILE* stream);
extern char *check_malloc();
extern FILE *check_fopen();
extern char *concatenate();

#endif
