/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "fileio.h"
#include "epwic.h"

void epwic_usage()
  { 
  printf("Usage: \nepwic infile [-o outfile] [-x xdim] [-y ydim] [-l levels] [-b outfile_size]\n"); 
  exit(-1);
  }

/* 
==========================================================================
parse_epwic_args() -- parse the arglist to the program, determining the 
input and output filenames, and the values of x_size, y_size, num_levels,
and compression_factor.  Read the image data from the input file into the 
float array, and open the output file for writing, checking for errors.
==========================================================================
*/

void parse_epwic_args (int argc,char *argv[],float **image,int *x_size, 
		       int *y_size, int *num_levels,long *outfile_size, 
		       FILE **outfile)
  {
  FILE *infile;
  char *filename;
  int arg_ptr=2;   /* first arg is requred infile name */

  /* fill in default values here, then override below if args are present */
  *x_size = *y_size = 256;
  *num_levels = -1;   /* indicates this is not set yet */
  *outfile_size = -1; /* Save the whole thing */

  if (argc < 2) epwic_usage();
  filename = argv[1];
  infile = check_fopen(filename, "r");

  filename = concatenate(argv[1], ".EPW");   /* default outfile */
  
  while (arg_ptr < argc)
    {
      if (!strcmp(argv[arg_ptr], "-o"))
	{
	  arg_ptr++;
	  filename = argv[arg_ptr];
	}
      else if (!strcmp(argv[arg_ptr], "-l"))
	{
	  arg_ptr++;
	  *num_levels = atoi(argv[arg_ptr]);
	}
      else if (!strcmp(argv[arg_ptr], "-b"))
	{
	  arg_ptr++;
	  *outfile_size = (long) atol(argv[arg_ptr]);
	}
      else if (!strcmp(argv[arg_ptr], "-x"))
	{
	  arg_ptr++;
	  *x_size = atoi(argv[arg_ptr]);
	}
      else if (!strcmp(argv[arg_ptr], "-y"))
	{
	  arg_ptr++;
	  *y_size = atoi(argv[arg_ptr]);
	}
      else { printf("Unrecognized argument: %s\n", argv[arg_ptr]); epwic_usage(); }
      arg_ptr++;
    }

  if (PGMStream(infile)){
    *image = (float *) ReadMatrixFromPGMStream(infile,x_size,y_size);
    if (*image == NULL){
      fprintf(stderr,"Unable to read PGM format\n");
      exit(-1);
    }
  } else {
    *image = (float *) check_malloc(*x_size * *y_size * sizeof(**image));
    read_byte_image(infile, *image, *x_size, *y_size);
  }
  fclose (infile);

  if ( *num_levels == -1 )  /* Choose num_levels automatically */
    {
      *num_levels = 1; /* initialize */
      while (( (*x_size % (1<<*num_levels)) == 0 ) &&
	     ( (*y_size % (1<<*num_levels)) == 0 ) &&
	     ( (*x_size / (1<<*num_levels)) >= FILTER_SIZE ) &&
	     ( (*y_size / (1<<*num_levels)) >= FILTER_SIZE ))
	{ *num_levels += 1; }
      *num_levels -= 1;
    }
  if ( *num_levels > MAX_LEVELS )
    { 
      printf("Error: cannot build pyramid to more than %d levels.\n", MAX_LEVELS); 
      exit(-1);
    }
  if (( *x_size > MAX_IMAGE_DIM ) ||
      ( *y_size > MAX_IMAGE_DIM ))
    {
      printf("Error: dimensions too big (%d,%d).\n", *x_size, *y_size);
      exit(-1);
    }
  if (( (*x_size % (1<<*num_levels)) != 0 ) ||
      ( (*y_size % (1<<*num_levels)) != 0 ))
    {
      printf ("Error: dimensions are not divisible by 2^num_levels (%d,%d).\n", 
	      *x_size, *y_size);
      exit(-1);
    }
  if (( (*x_size / (1<<(*num_levels-1))) < FILTER_SIZE ) ||
      ( (*y_size / (1<<(*num_levels-1))) < FILTER_SIZE ))
    {
      printf ("Error: dimensions (%d,%d) divided by 2^(num_levels-1) must be greater than %d.\n", 
	      *x_size, *y_size, FILTER_SIZE);
      exit(-1);
    }
  *outfile = check_fopen(filename, "w");
  }

void parse_pview_args(int argc, char *argv[], FILE **epwicfile, 
		      int *num_levels, int *x_size, int *y_size)
  {
  char id[4];
  char version;
  char *filebuf;
  int i;


  if (argc > 1){
    *epwicfile = fopen(argv[1],"r");
    if (*epwicfile == NULL){
      printf("Unable to open %s for input\n",argv[1]);
      exit(0);
    }
  } else {
    *epwicfile = stdin;
  }

  /* Give the inputfile a large buffer of 16 K*/
  filebuf = (char *) malloc(16384);
  setvbuf(*epwicfile,filebuf,_IOFBF,16384);

  for (i=0;i<strlen(EPWIC_ID_TAG);i++){
    read_byte(id[i], *epwicfile);
  }
  id[strlen(EPWIC_ID_TAG)] = '\0';
  if (strcmp(id,EPWIC_ID_TAG)){ 
    fprintf(stderr,"Error: this file is not an EPWIC file!\n");
    exit(-1); 
  }

  read_byte(version, *epwicfile);
  if (version != VERSION_TAG){
    fprintf(stderr,"Warning: EPWIC version does not match file. Trying anyway...\n");
  }

  read_byte(temp_byte, *epwicfile); *num_levels = (int) temp_byte;
  read_short(temp_short, *epwicfile); *x_size = (int) temp_short;
  read_short(temp_short, *epwicfile); *y_size = (int) temp_short;

}

/* unepwic */
void parse_unepwic_args(int argc, char *argv[], FILE **epwicfile, 
			int *num_levels, int *x_size, int *y_size, 
			FILE **outfile, unsigned long *bytes_to_read)
  {
  char *filename=NULL;
  int arg_ptr=1;   /* first arg is requred infile name */
  char id[4];
  char version;
  int i;


  *bytes_to_read = -1;
  *outfile = stdout;
  *epwicfile = stdin;

  if ((argc >= 2) && (strcmp(argv[1],"-h") == 0)) {
     printf("usage:\n");
     printf("   %s [filename] [-o outfile] [-b bytes_to_read]\n",argv[0]);
     exit(0);
  }

  if (argc >= 2 && argv[1][0] != '-'){
     filename = argv[1];
     *epwicfile = check_fopen(filename, "r");
     arg_ptr++;
  }
  
  while (arg_ptr < argc)
    {
      if (!strcmp(argv[arg_ptr], "-o"))
	{
	  arg_ptr++;
	  filename = argv[arg_ptr];
          *outfile = check_fopen(filename,"w");
	}
      else if (!strcmp(argv[arg_ptr], "-b"))
	{
	  arg_ptr++;
	  *bytes_to_read = (unsigned long) atol(argv[arg_ptr]);
	}
      else 
	{ 
	  fprintf(stderr,"Unrecognized argument: %s\n", argv[arg_ptr]);
	  fprintf(stderr,"usage:\n");
	  fprintf(stderr,"   %s [filename] [-o outfile] [-b bytes_to_read]\n",argv[0]);
	  exit(-1);
	}
        arg_ptr++;
    }

  for (i=0;i<strlen(EPWIC_ID_TAG);i++){
    read_byte(id[i], *epwicfile);
  }
  id[strlen(EPWIC_ID_TAG)] = '\0';
  if (strcmp(id,EPWIC_ID_TAG)){ 
    fprintf(stderr,"Error: this file is not an EPWIC file!\n");
    exit(-1); 
  }

  read_byte(version, *epwicfile);
  if (version != VERSION_TAG){
    fprintf(stderr,"Warning: EPWIC version does not match file. Trying anyway...\n");
  }

  read_byte(temp_byte, *epwicfile); *num_levels = (int) temp_byte;
  read_short(temp_short, *epwicfile); *x_size = (int) temp_short;
  read_short(temp_short, *epwicfile); *y_size = (int) temp_short;
}
