/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "epwic.h"  /* For log2() */
#include "pyramid_util.h"
#include "matrix.h"  /* matrix and least squares functions*/

/* These define the number of intervals between 0 and 512 for
   the first order distribution.  XSTEP * (XSIZE-1) = 512. */

#define XSIZE     513
#define XSTEP     1

/* Note: the order of the subbands is horizontal, vertical,
   and then diagonal. */

/* This array is the function inv(log10(gamma(e)*gamma(5e)/(gamma(3e)^2)))
   evaluated from e = 0.4 to 2.4 in increments of 0.05 */

double p_from_kurt_func[] = 
 {0.363627, 0.452834, 0.539309, 0.624062, 0.707648, 0.790398,
  0.872531, 0.954191, 1.03548, 1.11648, 1.19724, 1.2778, 1.3582, 1.43846,
  1.5186, 1.59864, 1.6786, 1.75848, 1.8383, 1.91806, 1.99777, 2.07743,
  2.15705, 2.23664, 2.31619, 2.39572, 2.47521, 2.55468, 2.63413, 2.71356,
  2.79298, 2.87237, 2.95175, 3.03111, 3.11046, 3.1898, 3.26913, 3.34845,
  3.42775, 3.50705, 3.58634};

double p_from_kurt(double kurtosis){
  double logkurt = log(kurtosis)/log(10);
  int entry;
  double fract;
  double e;

  if (logkurt <= 0.4){
    fprintf(stderr,"Warning:  kurtosis of %f too small\n",kurtosis);
    e = p_from_kurt_func[0];
  } else if (logkurt >= 2.4) {
    fprintf(stderr,"Warning:  kurtosis of %f too big\n",kurtosis);
    e = p_from_kurt_func[40];
  } else {
    entry = (int) floor((logkurt - 0.4) / 0.05);
    fract = (logkurt - (((double)entry) * 0.05 + 0.4)) / 0.05;
    e = (p_from_kurt_func[entry] * (1-fract) + 
	 p_from_kurt_func[entry+1] * fract);
  }

  return (1.0/e);
}

/* This array is the function log10(gamma(3e)/gamma(e))
   evaluated from e = 0.5 to 3.5 in increments of 0.05 */

double s_from_var_func[] = 
 {-0.30103, -0.254176, -0.203822, -0.150212, -0.0935598, -0.0340527,
   0.0281421, 0.0928755, 0.160013, 0.229435, 0.30103, 0.374699, 0.450351,
   0.527901, 0.607273, 0.688395, 0.771201, 0.855629, 0.941624, 1.02913,
   1.1181, 1.20848, 1.30024, 1.39333, 1.48771, 1.58334, 1.68019, 1.77824,
   1.87743, 1.97776, 2.07918, 2.18168, 2.28522, 2.38978, 2.49534, 2.60188,
   2.70937, 2.8178, 2.92714, 3.03738, 3.1485, 3.26048, 3.3733, 3.48695,
   3.60142, 3.71668, 3.83273, 3.94955, 4.06712, 4.18544, 4.30449, 4.42426,
   4.54474, 4.66591, 4.78777, 4.91031, 5.03351, 5.15736, 5.28186, 5.407,
   5.53276};

double s_from_var(double variance,double p){
  double e = 1/p;
  int entry;
  double fract;

  if (e <= 0.5){
    fprintf(stderr,"Warning:  p of %f too big\n",p);
    e = s_from_var_func[0];
  } else if (e >= 3.5) {
    fprintf(stderr,"Warning:  p of %f too small\n",p);
    e = s_from_var_func[60];
  } else {
    entry = (int) floor((e - 0.5) / 0.05);
    fract = (e - (((double)entry) * 0.05 + 0.5)) / 0.05;
    e = (s_from_var_func[entry] * (1-fract) + 
	 s_from_var_func[entry+1] * fract);
  }

  return (sqrt(variance * pow(10,-e)));
}

/* Calculate the two parameters (s and p) which can be used to estimate
   the first order pdf of the subband */ 

void estimate_pdf(Pyramid pyr, int subband, double *s, double *p){
  int i;
  double variance, kurtosis;
  double value;
  int child_x, child_y;

  /* Initialize variance and kurtosis */
  variance = 0;
  kurtosis = 0;

  /* Calculate variance, and temporarily calculate the fourth moment
     in the variable kurtosis */
  for(i=0; i<pyrSubSize(pyr,subband); i++){

    child_x = i % pyrSubXSize(pyr,subband);
    child_y = i / pyrSubXSize(pyr,subband);
   
    value =  (double)(pyrCoefCoord(pyr,subband,child_x,child_y)
                     & COEF_ABS_MASK);

    value = value / (double) (1<<MAX_NUM_BITS) * 512.0;

    value = value * value;

    variance += value;
    kurtosis += value * value;
  }
  variance = variance / ((double) pyrSubSize(pyr,subband));
  kurtosis = kurtosis / ((double) pyrSubSize(pyr,subband));   
         /* fourth moment */
  kurtosis = kurtosis / (variance * variance);  /* kurtosis */

  /* Now calculate the values */
  (*p) = p_from_kurt(kurtosis);
  (*s) = s_from_var(variance,(*p));

}

double **PriorCDF;   /* Used to hold calculated proir CDF's */

/* Mallocs the needed memory for the PriorCDF */

void init_prior(int num_levels){
  int i;

  PriorCDF =(double **) malloc(((num_levels * 3)+1) * sizeof(double *));
  for(i=0;i<(num_levels *3)+1;i++){
    PriorCDF[i] = NULL;
  }
}

/* This states if we have a prior CDF for a particular subband yet */

int exist_prior(int subband){
  return(PriorCDF[subband] != NULL);
}

/* This normalizes a matrix by column (so the sum of each column is 1) */

void norm_sum_c(double * m, int x_size, int y_size){
  double sum;
  int i,j;

  for(i=0;i<x_size;i++){
    sum = 0;
    for(j=0;j<y_size;j++){
      sum += m[i+j*x_size];
    }
    for(j=0;j<y_size;j++){
      m[i+j*x_size] = m[i+j*x_size] / sum;
    }
  }
}

/* Given a subband, an s, and a p, this calculates the estimate of the
   prior CDF */

void calc_prior(int subband, double s, double p){
  int i;
  int j;
 
  PriorCDF[subband] = (double *) malloc( XSIZE * sizeof(double));

  /* First we calculate the probability density function */

  for (i=0;i<XSIZE;i++){
    PriorCDF[subband][i] = 0;
    for(j=i*XSTEP;j<(i+1)*XSTEP;j++){
      PriorCDF[subband][i] += exp(-pow(j/s,p));
    }
    PriorCDF[subband][i] = PriorCDF[subband][i] / (double)(XSTEP);
  }

  PriorCDF[subband][0] *= 0.5; 

  /* Then we normalize to 1 */

  norm_sum_c(PriorCDF[subband],1,XSIZE);

  /* Finally we sum the PDF to make a CDF */

  for (i=1;i<XSIZE;i++){
    PriorCDF[subband][i] += PriorCDF[subband][i-1];
  }

}

/* Returns value of prior cdf for a given value */
/* If used, we will want to rescale the values between 0 and 512.0 */

double get_prior_CDF(int subband,int value){
  int v;
  double prob;
  double f;

  value = (int) (value / (double) (1<<MAX_NUM_BITS) * 512.0);
  v = value / XSTEP;
  f = (value / (double)(XSTEP)) - v;

  /* Use linear interpolation to find probability */
  prob =  PriorCDF[subband][v] * (1-f);
  if (f > 0) prob += PriorCDF[subband][v+1] * f;

  return(prob);
}

Coefficient EstTable[512];
int EstReadToBit;
int UseDefaultEst;
Coefficient DefaultEst = 0;


/* Assume 9 bits accuracy for estimation */
void init_estimation_fraction(int subband, int read_to_bit){
  EstReadToBit = read_to_bit;
  UseDefaultEst = 1;

  if (DefaultEst == 0){
    DefaultEst = floor(0.32 * pow(2.0,sizeof(Coefficient)*8));
  }
  if (MAX_NUM_BITS - read_to_bit >= 9){ /* Gone beyond what we can estimate */
    return;
  }
  if (exist_prior(subband)){
    double PDF[512];
    int i;
    int j;
    int est_len;
    int sum_size;
    int val;
    double avg;
    double sum_prob;
    
    UseDefaultEst = 0;
    PDF[0] = get_prior_CDF(subband,0);
    for(i=1;i<512;i++){ /* Calculate probability density function */
      PDF[i]=get_prior_CDF(subband,i) - get_prior_CDF(subband,i-1);
    }
    est_len = 1<<(MAX_NUM_BITS - read_to_bit);
    sum_size = 512/est_len;
    
    for(i=0;i<est_len;i++){
      avg=0;
      sum_prob = 0;
      for(j=0;j<sum_size;j++){
	val = i*sum_size + j;
	avg += ((double) val) * PDF[val];
	sum_prob += PDF[val];
      }
      if (sum_prob > 0){
	avg = avg / sum_prob;
      } else {
	avg = i*sum_size;
      }
      EstTable[i] = (Coefficient)(avg - i*sum_size);
    }
  }
}

Coefficient get_estimation_fraction(Coefficient coef){
  if (UseDefaultEst){
    return ((DefaultEst) >>
	    ((sizeof(Coefficient)*8) - EstReadToBit));
  } else {
#if (MAX_NUM_BITS > 9)
      return (EstTable[(int)coef >> (MAX_NUM_BITS - 9)]);
#else
      return (EstTable[(int)coef << (9 - MAX_NUM_BITS)]);
#endif
  }
    /*
    if (MAX_NUM_BITS>9){
      return (EstTable[(int)coef >> (MAX_NUM_BITS - 9)]);
#else
    } else {
      return (EstTable[(int)coef << (9 - MAX_NUM_BITS)]);
    }
    */
}


