/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include "arith.h"
#include "epwic.h"
#include "quantize.h"
#include "collapse_pyr.h"
#include "utilities.h"
#include "parse_args.h"
#include "xfb_byte.h"
#include "run_length.h"
#include "arith.h"
#include "prob_dist.h"
#include "pyramid_util.h"


#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_END 2
#endif

#define _MAX(a,b) (a<b)?b:a
#define MIN_WIDTH  256

void write_img_line(bytes_read, x_size,y_size)
  int bytes_read;
  int x_size, y_size;
  {
  char message[80];  

  sprintf(message,"Image: %d bytes (%.2f bits/pixel)",
	  bytes_read,
	  ((float) (8 * bytes_read)) / ((float)(x_size * y_size)));
  DrawBlackBox(1,0,y_size,_MAX(x_size,MIN_WIDTH),17);
  WriteText(1,message,2,y_size+15);
  }

#define DSP_DONE 0
#define DSP_PAUSED 1
#define DSP_READING 2
#define DSP_DISPLAYING 3 

/* state = {0,1,2}.  0=> done.  1=> paused.  2=> reading data. 3=> redisplaying.*/
void write_pause_line(state, bytes_read, x_size,y_size)
  int state;
  int bytes_read;
  int x_size, y_size;
  {
  char message[80];  

  if (state == DSP_DONE) 
    sprintf(message,"Image Complete");
  else if (state == DSP_PAUSED)
    sprintf(message,"Left Button: Continue");
  else if (state == DSP_READING)
    sprintf(message,"Left Button: Pause at %d bytes", bytes_read);
  else /* DSP_DISPLAYING */
    sprintf(message,"Redisplaying ...");
  DrawBlackBox(1,0,y_size+17,_MAX(x_size,MIN_WIDTH),17);
  WriteText(1,message,2,y_size+32);
  }

void write_quit_line(x_size,y_size)
  int x_size, y_size;
  {
  WriteText(1,"Right Button: Quit",2,y_size+49);
  }

void display_pyramid(q_pyr,x_size,y_size,num_levels,bin_size,scale_factor,bytes_read)
Coefficient *q_pyr;
int x_size;
int y_size;
int num_levels;
BinValueType *bin_size;
float scale_factor;
int bytes_read;
{
    int *pyr;
    int *result;
    Byte *image;
    int i;
    double dtemp;
    double min;

    write_pause_line(DSP_DISPLAYING,bytes_read,x_size,y_size);
    WindowExposure(1);  /* Hack to flush the Display message */

    /* -- Unquantize the pyramid -- */
    pyr = (int *) check_malloc( x_size*y_size*sizeof(*pyr) );
    unquantize_pyr(q_pyr, pyr, (x_size*y_size), num_levels, bin_size);

    /* -- Collapse pyramid, and divide by scale_factor -- */
    result = (int *) check_malloc(x_size*y_size*sizeof(*result));
    image = (Byte *) check_malloc(x_size*y_size*sizeof(*image));

    collapse_pyr(pyr, result, x_size, y_size, num_levels);
    min = 0.0;

    for (i=0; i<x_size*y_size; i++) {
        dtemp = (result[i] - min) / scale_factor;
        if (dtemp < 0) image[i] = 0;              /* clip below */
        else if (dtemp > 255) image[i] = 255;     /* clip above */
        else image[i] = (int) (dtemp + 0.5);      /* round off  */
    }

    check_free((char *) result);
    DrawImage8(1,image,0,0,x_size,y_size);
    write_img_line(bytes_read, x_size,y_size);

    check_free((char *) image);
}

/* When passed a timb struct from the past, it calculates the number of 
   milliseconds from now.  This can be used as a timer */

long msec_from_now(struct timeb *old_time){
   struct timeb new_time;
   long millisec_elapsed;

   ftime(&new_time);
   millisec_elapsed = (long)(new_time.millitm - old_time->millitm) +
			((new_time.time - old_time->time)*1000);
   return millisec_elapsed;
}

/* While we read in the data, we can be in two states:
        1) We need to read in the next type of data (or end of file)
        2) We are in the process of reading in the level of a pyramid
*/

#define GET_DATA_TYPE 1
#define READ_LEVEL    2
#define DONE          3

#define RLE_DECODE    1
#define ARITH_DECODE  2

int main(argc, argv)
  int argc;
  char *argv[];
  {
  FILE *infile;
  Coefficient *q_pyr;
  Pyramid pyr;
  int x_size, y_size, num_levels;
  double scale_factor;
  Byte the_tag;
  int i;
  int next_symbol;
  int *read_to_bit;
  int subband,im_size = 0;
  int im_offset,pyr_offset;
  Coefficient est_mask;
  int state;
  int bytes_read;
  int old_bytes_read;
  int bin_bytes_read;
  int message_time;
  int read_sign;

  int read_since_display;  /* 0 means nothing read since last display
                              1 - Beginning of block read
                              2 - End of block read (can redisplay) */
  int infd; /* File descriptor of stream infile */
  int val;  /* File descriptor flags of infd */
  int testchar; /* used to check if buffer is empty */

  arithcode acode;
  int decode_type = 0;
  unsigned char temp_char;

  scale_factor = 128.0;  /* Hard coded right now */

  /* --- Parse arglist, Reading the data from the input file, unpacking it, 
     and decoding it, and putting it into the quantized pyramid --- */
  parse_pview_args(argc, argv, &infile, &num_levels, &x_size, &y_size);

  /* Initialize pyramid datastructure */

  q_pyr = (Coefficient *) check_malloc( x_size*y_size*sizeof(*q_pyr) );
  bzero((char*) q_pyr, x_size * y_size * sizeof(Coefficient));
  pyr = pyrNewPyramid(q_pyr, x_size, y_size, num_levels);
  
  init_prior(num_levels);


  /* Init arithmetic coding */
  acode.ilob=(unsigned long *)lvector(1,NWK);
  acode.iupb=(unsigned long *)lvector(1,NWK);
  acode.codep = (unsigned char *)malloc(NWK * sizeof(unsigned char));
  acode.lcode = NWK;
  arcmak(256,&acode);

  infd = fileno(infile);  /* Get the file descriptor of the stream */

  /* BIN INFO NOW READ AT LAST MOMENT
      read_byte(the_tag, infile);
      if (the_tag != (Byte)BIN_INFO_TAG) {
        fprintf(stderr,"Expecting BIN_INFO first...\n");
        exit(-1);
      }
      ForEachSubband(pyr,i){
          read_short(pyrSubBinSize(pyr,i), infile);
      }
  */

  /* Init bin sizes to 0 */
  ForEachSubband(pyr,i){
    pyrSubBinSize(pyr,i) = 0;
  }

  /* -- now to start the whole mess -- */


  read_to_bit = (int *) check_malloc((3*num_levels+1)*sizeof(*read_to_bit));
  for(i=0;i<3*num_levels+1;i++){
	read_to_bit[i] = (int) ceil(log2((double)MAX_ABS_COEFF))-1;
  }

  state = GET_DATA_TYPE;
  subband = 0;     /* Level of the pyrmid we are reading */
  pyr_offset = 0; /* Location of the level in the array */
  im_offset = 0;  /* Location in the level we are at */
  bytes_read = 9; /* Calculate size of header */
  old_bytes_read = bytes_read;
  bin_bytes_read = 0;
  message_time = 256;
  read_since_display = 0; /* Mark beginning of block as read */
  read_sign = 0;

  InitX(1,argv[1],20,20,_MAX(x_size,MIN_WIDTH),y_size+52);
  write_pause_line(DSP_READING, bytes_read, x_size, y_size);
  write_quit_line(x_size,y_size);
  while (WindowExposure(1)) {};  /* slurp up the initial exposure event */

  while(state != DONE){

    i = ButtonPressed(1);
    if (i==1){  /* left button */

      if (read_since_display == 2){
	display_pyramid(q_pyr,x_size,y_size,num_levels,
			pyr->sub_bin_size,scale_factor,bytes_read);
	read_since_display = 0;
      }

      write_pause_line(DSP_PAUSED, bytes_read, x_size, y_size);

      i = ButtonPressed(1);  /* Hack to flush the Display message */
      while(i != 1){ /* Pause until button pressed again */
	i=ButtonPressed(1);
	if (i==3) exit(0);
	if (WindowExposure(1)){
	  display_pyramid(q_pyr,x_size,y_size,num_levels,
			  pyr->sub_bin_size,scale_factor,bytes_read);
	  write_pause_line(DSP_PAUSED, bytes_read, x_size,y_size);	  
	  write_quit_line(x_size,y_size);
	}
      }
      write_pause_line(DSP_READING, bytes_read, x_size,y_size);
    } else if (i==3) exit(0);
    
    if (WindowExposure(1)){
      read_since_display = 0;
      display_pyramid(q_pyr,x_size,y_size,num_levels,
		      pyr->sub_bin_size,scale_factor,bytes_read);
      write_pause_line(DSP_READING, bytes_read, x_size,y_size);

      write_quit_line(x_size,y_size);
    }
    
    if (bytes_read > message_time){
      message_time += 256;
      write_pause_line(DSP_READING, bytes_read, x_size,y_size);
    }
    
    /* Check to see if we've read enough data and the buffer is empty */
    if (read_since_display == 2){
      val = fcntl(infd, F_GETFL);
      val |= O_NONBLOCK;
      fcntl(infd, F_SETFL, val);  /* Set non blocking */
      errno = 0;  /* Clear errno */
      if ((testchar = fgetc(infile)) == EOF){
	if (errno == EAGAIN){
          read_since_display = 0;
	  display_pyramid(q_pyr,x_size,y_size,num_levels,
			  pyr->sub_bin_size,scale_factor,bytes_read);
	} else if (feof(infile)){
          state = DONE;
	} else {
          printf("We got errno %d\n",errno);
	}
      } else {
        ungetc(testchar,infile);
      }
	
      val &= ~ O_NONBLOCK;
      fcntl(infd, F_SETFL, val);
      
    }
       

    if (state == GET_DATA_TYPE){

      bytes_read += arith_decode_file_byte(infile,&the_tag,&acode);
      old_bytes_read = bytes_read;
      bin_bytes_read = 0;

      if (the_tag == (Byte) DONE_DATA_TAG){ 
	state = DONE;
      } else if (the_tag & (Byte) DATA_MASK){
        if (read_since_display < 2) read_since_display++;

        state = READ_LEVEL;
        subband = the_tag & 0x1F;

        /* Get Bin Info if needed */
        if(pyrSubBinSize(pyr,subband) == 0){
          bin_bytes_read += arith_decode_file_byte(infile,&temp_char,&acode);
          pyrSubBinSize(pyr,subband) = ((unsigned short)temp_char) << 8;
          bin_bytes_read += arith_decode_file_byte(infile,&temp_char,&acode);
          pyrSubBinSize(pyr,subband) += ((unsigned short)temp_char);
        }

	im_offset = 0;
        if (subband==0){
	  im_size = (x_size * y_size) >> (2*num_levels);
	  pyr_offset = 0;
        } else {
	  im_size = (x_size * y_size) >> (2*(num_levels-((subband-1)/3)));
            pyr_offset = (((subband-1)%3)+1) * im_size;
        }

	if ((the_tag & (Byte) DATA_MASK) == (Byte) RAW_DATA_TAG){
	  rle_init_decode(infile,0,0,&acode); /* raw */
	  decode_type = RLE_DECODE;
	} else if ((the_tag & (Byte) DATA_MASK) == (Byte) RLE_TYPE_ONE){
	  rle_init_decode(infile,1,2,&acode); /* 8,16 */
	  decode_type = RLE_DECODE;
	} else if ((the_tag & (Byte) DATA_MASK) == (Byte) RLE_TYPE_TWO){
	  rle_init_decode(infile,2,2,&acode); /* 16,16 */
	  decode_type = RLE_DECODE;
	} else { /* arith decode*/
	  arith_bitplane_init_decode(infile,pyr,subband,
				     read_to_bit,&acode); 
	  decode_type = ARITH_DECODE;
	}

	/* Initiate estimation lookup table */
	init_estimation_fraction(subband,read_to_bit[subband]);

      } else {
        fprintf(stderr,"Unknown data tag... ignoring\n");
      } 
    } else if (state == READ_LEVEL){
      if (decode_type == RLE_DECODE){
	rle_read_bit_by_byte(&next_symbol);
	bytes_read = old_bytes_read + rle_bytes_read() + bin_bytes_read;
      } else {
	next_symbol = arith_bitplane_read_bit();
	bytes_read = old_bytes_read + arith_bitplane_bytes_read() +
                     bin_bytes_read;
      }

      if (read_sign){	  /* Store the sign bit */

	read_sign = 0;
	pyrCoefOffset(pyr,subband,im_offset) |=
	  ((next_symbol)&1)*COEF_SIGN_BIT;

	/* Here we make an estimate about the rest of the number */
	if ((read_to_bit[subband] > 0) && 
	    (pyrCoefOffset(pyr,subband,im_offset) & ~((Coefficient) 
				   ((1<<(read_to_bit[subband])) -1)))){

	  if (read_to_bit[subband] < MAX_NUM_BITS){
	    /* Remove old estimation */
	    pyrCoefOffset(pyr,subband,im_offset) &=
	      ~((Coefficient) (1<<(read_to_bit[subband])) - 1);
	  }
	  est_mask = get_estimation_fraction(
		     (pyrCoefOffset(pyr,subband,im_offset) & ~(COEF_SIGN_BIT))
		     >> read_to_bit[subband]);
				   
	  pyrCoefOffset(pyr,subband,im_offset) |= est_mask;
	}
	im_offset ++;  /* With the sign read, we can move on */

      } else if (pyrCoefOffset(pyr,subband,im_offset) >> 
		 (read_to_bit[subband]+1)){
	/* significant bit sent */

	/* First remove the value at the bit to which we are
	   storing the new value */
	pyrCoefOffset(pyr,subband,im_offset) &=
                      ~((Coefficient) 1<<(read_to_bit[subband]));

	/* Store the new bit read */
	pyrCoefOffset(pyr,subband,im_offset) |= 
	    ((next_symbol)&1)<<(read_to_bit[subband]);

	/* Here we make an estimate about the rest of the number */
	if ((read_to_bit[subband] > 0) && 
	    (pyrCoefOffset(pyr,subband,im_offset) & ~((Coefficient) 
				   ((1<<(read_to_bit[subband])) -1)))){

	  if (read_to_bit[subband] < MAX_NUM_BITS){
	    /* Remove old estimation */
	    pyrCoefOffset(pyr,subband,im_offset) &=
	      ~((Coefficient) (1<<(read_to_bit[subband])) - 1);
	  }
	  est_mask = get_estimation_fraction(
		     (pyrCoefOffset(pyr,subband,im_offset) & ~(COEF_SIGN_BIT))
		     >> read_to_bit[subband]);
				   
	  pyrCoefOffset(pyr,subband,im_offset) |= est_mask;
	}
	im_offset ++;
      } else { /* significant bit not sent */
	if ((next_symbol)&1) { /* reached significant bit */
	  pyrCoefOffset(pyr,subband,im_offset) |= 
	      ((next_symbol)&1)<<(read_to_bit[subband]);
	  read_sign = 1; /* Read sign bit next */
	  /* Do NOT increment im_offset since we need to read sign bit */
	} else { /* not reached significant bit */
	  im_offset ++;
	}
      }
      if (im_offset >= im_size){
	state = GET_DATA_TYPE;
	read_to_bit[subband]--;
      }
    }
  }
  
  display_pyramid(q_pyr,x_size,y_size,num_levels,pyr->sub_bin_size,
		  scale_factor,bytes_read);
  write_pause_line(DSP_DONE, bytes_read, x_size,y_size);
  
  fclose(infile);
  
  i = 0;
  while(i != 3){ /* Pause until right button */
    i=ButtonPressed(1);
    if (WindowExposure(1)){
      display_pyramid(q_pyr,x_size,y_size,num_levels,pyr->sub_bin_size,
		      scale_factor,bytes_read);
      write_pause_line(DSP_DONE, bytes_read, x_size,y_size);
      write_quit_line(x_size,y_size);
    }
  }

  return(1);  
}


