/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <math.h>
#include <stdlib.h>
#include "epwic.h"
#include "pyramid_util.h"

#define bit_on(val,mask)  (val |= (mask))
#define bit_off(val,mask) (val &= (~(mask)))

static Coefficient sign_bit = 1<<(8*sizeof(Coefficient)-1);

/* 
==========================================================================
quantize_image() -- Quantizes float_im and puts integer index values
into int_im.  The index values are simply round(val/binsize). 
==========================================================================
*/

void quantize_image (float_im, q_im, im_size, bin_size, range)
  register float *float_im;
  register BinValueType *bin_size;
  register Coefficient *q_im;
  int im_size;
  double *range;
  {
  register int i;
  register double ftemp, max_abs;
  Coefficient q_val;

  max_abs = 0.0;
 
  /* find maximum abs of image */
  for (i=0; i<im_size; i++)
      {
      ftemp = float_im[i];
      if (ftemp < 0.0)       ftemp = -ftemp;
      if (ftemp > max_abs)  max_abs = ftemp;
      }
 
  (*bin_size) = 0; /* Force binsize to be too small */ 
  /* Check for too many bins */
  if ( max_abs >= ((MAX_ABS_COEFF-2)*(*bin_size)/2) )    
      { 
      i = *bin_size;
      *bin_size = (BinValueType) ceil( max_abs/(MAX_ABS_COEFF-2));
      /*
        printf("QUANTIZE: bin_size %d is too small. Changed to %d.\n",
	       i, (*bin_size));
      */
      }

  for (i=0; i<im_size; i++) {
      ftemp = float_im[i] / *bin_size;
      if (ftemp < 0.0) ftemp = -ftemp;
      q_val = (Coefficient) (ftemp+0.5);
      if (q_val > MAX_ABS_COEFF){
	printf("TOO BIG\n");
	q_val = MAX_ABS_COEFF;
      }
      if (float_im[i] < 0.0) bit_on(q_val,sign_bit);
      q_im[i] = q_val;
      }

  (*range) = max_abs;
  }

			
/*
==========================================================================
quantize_pyr() -- quantizes the sub-band images in the pyramid.  The
pyramid should be stored in the order: L(N) V(N) H(N) D(N) V(N-1)
H(N-1) D(N-1) V(N-2) ... D(1), where L=Lowpass, V=Vertical,
H=Horizontal, D=Diagonal and the number in paretheses is the level
number.  Compression_factor is a user parameter which determines how
much compression is desired.  The quantize_pyr routine must determine
the binsizes to be used.  For now, we use the same binsize for each
band, decreasing by a constant factor at each level.  CLEARLY NON-OPTIMAL!
==========================================================================
*/

void quantize_pyr (float *float_pyr, Coefficient *q_pyr, int pyr_size, 
		   int num_levels, short scale_factor, 
		   BinValueType *bin_size, int *salience, int *bottom_bit)
  {
  int i;
  int level, pyr_offset;
  double range;
  int maxsalience;
  float the_salience;
  float the_bottom_bit;
  int im_size;
  float level_mult = SALIENCE_LEVEL_MULT;

  while (level_mult < 0.0){
    printf("Enter Level Mult: ");
    scanf("%f",&level_mult);
    printf("Level Mult = %f\n",level_mult);
  }

  salience[0] = 0;

  for (i=0;i<(3*num_levels+1);i++){
    if (i==0){
		level = num_levels;
        im_size = (pyr_size) >> (2*num_levels);
        pyr_offset = 0;
    } else {
	level = num_levels-((i-1)/3);
        im_size = (pyr_size) >> (2*(num_levels-((i-1)/3)));
        pyr_offset = (((i-1)%3)+1) * im_size;
    }

    quantize_image(float_pyr+pyr_offset, q_pyr+pyr_offset, im_size, 
		       (bin_size+i),&range);

    if (range < 0.0001) range = 0.0001; /* Make sure we don't take log of 0 */

    /* The bottom bit is the range divided by the "standard" 
       bin size (which is scale factor / 4 * 2 ^ (- level) ) */

/*
    the_bottom_bit = log2(4.0 * range / (float) scale_factor)-
                         ((i==0)?0:(num_levels - level+1));

    Don't do level compensation
*/
    the_bottom_bit = log2(4.0 * range / (float) scale_factor);
    the_bottom_bit =  (the_bottom_bit<0.0)?
                            (the_bottom_bit-0.5):(the_bottom_bit+0.5);
    bottom_bit[i] = (int) ceil(log2((double)MAX_ABS_COEFF)) -
                        (int) the_bottom_bit - 1;


/* Zero bottom bit*/
bottom_bit[i] = (int) 0;

    /* printf("%d (%d) -> %f %d\n",i,((i==0)?0:(num_levels - level+1)),
                 the_bottom_bit, bottom_bit[i]); */

    the_salience = log2(range) + (level_mult*level);

    salience[i] = (int)	((the_salience<0.0)?
                             (the_salience-0.5):(the_salience+0.5));
  }
  
  for(maxsalience = salience[0], i=1; i<(3*num_levels+1); i++) {
    if (maxsalience < salience[i]) maxsalience = salience[i];
  }
  for(i=0;i<(3*num_levels+1);i++){
	salience[i] = salience[i] - maxsalience;
  }
}
  
/* 
==========================================================================
unquantize_image() -- Take integer index image and multiply by
  binsize.  Returns a (signed) int image.
==========================================================================
*/

void unquantize_image(q_im, res, im_size, bin_size)
  register Coefficient *q_im;
  register int *res, im_size;
  register BinValueType bin_size;
  {
  register int i;

  for (i=0; i<im_size; i++)  
      {
      if (q_im[i] & sign_bit)
	  res[i] = - ((int) ((q_im[i] & (~sign_bit)) * bin_size));
      else
	  res[i] = (int) (q_im[i] * bin_size);
      }
  }


/*
==========================================================================
unquantize_pyr() -- multiplies the values of q_pyr by the bin_size.
==========================================================================
*/

void unquantize_pyr(Coefficient *q_pyr, int *pyr, int im_size, int num_levels, 
		    BinValueType *bin_size)
  {
  int level, bin_index, im_num, pyr_offset = 0;

  for (im_size>>=(num_levels*2), level=num_levels, bin_index=0;
       level > 0;
       level--, im_size<<=2)
    for (im_num=((level == num_levels) ? 0 : 1); 
	 im_num < 4;
	 im_num++, pyr_offset+=im_size, bin_index++)
	{
	/* printf ("Unquantizing level %d, image %d\n", level, im_num); */
	unquantize_image((q_pyr+pyr_offset), (pyr+pyr_offset), im_size, 
			 bin_size[bin_index]);
	}
  }

/* ========================================================================== */
/* Reorder the bits such that the sign bit comes just after the MSB .
   THis gives longer runs of zeros... */
void embed_sign_bit(in_array,out_array,sz)
  Coefficient *in_array, *out_array;
  int sz;
  {
  register Coefficient val, mask, prev_mask, sign;
  unsigned int i;
  
  for (i=0; i<sz; i++, in_array++, out_array++)
      {

      sign = sign_bit & (*in_array);     /* capture original sign bit */
      val = (*in_array) & (~sign_bit);   /* turn off orig sign bit */

      for (prev_mask=sign_bit, mask=sign_bit>>1; 
	   (mask > 0);
	   prev_mask=mask, mask>>=1)
	  {
	  if (mask & (*in_array))
	      {
	      bit_on(val,prev_mask);  /* prev bit becomes MSB */
	      if (sign) bit_on(val,mask); /* this bit becomes sign */
	      else bit_off(val,mask);    
	      mask = 0;		/* flag: done with this array elt */
	      }
	  }
      *out_array = val;
      }
  }

void unembed_sign_bit(in_array,out_array,sz)
  Coefficient *in_array, *out_array;
  int sz;
  {
  register Coefficient val, mask, prev_mask;
  unsigned int i;
 
  for (i=0; i<sz; i++, in_array++, out_array++)
      {
      val = (*in_array);

      for (prev_mask=sign_bit, mask=sign_bit>>1; 
	   mask != 0;
	   prev_mask=mask, mask>>=1)
	  {
	  if (prev_mask & val)
	      {
	      bit_off(val,prev_mask); /* turn off prev MSB */
	      if (val & mask)           /* if negative, */
	        bit_on(val,sign_bit);  /* set new sign bit */
	      bit_on(val,mask);    /* turn on new MSB */
	      mask = 0;		   /* flag: done with this array elt */
	      }
	  }
      *out_array = val;
      }
  }

/*
print_binary(val)
  Coefficient val;
  {
  int i;
  for (i=8*sizeof(Coefficient) - 1;i>=0;i--)
      {
      printf("%d",val>>i & 1);
      }
  }

main(argc, argv)
  int argc;
  char *argv[];
  {
  int i, pos, neg, imsz=65536;
  FILE *fp;
  Coefficient *q_im;
  float *image;
  double range;
  BinValueType bin_size;
  int *res;

  if (argc < 3) 
    { printf("Usage: quantize filename binsize\n"); exit(-1); }

  image = (float *) check_malloc(256*256*sizeof(*image));
  q_im = (Coefficient *) check_malloc(256*256*sizeof(*q_im));
  res = (int *) check_malloc(256*256*sizeof(*res));

  fp = check_fopen(argv[1], "r");
  read_byte_image(fp, image, 256, 256);
  fclose (fp);

  bin_size = atoi(argv[2]);

  for (i=0; i< imsz; i++)
      { 
      image[i] -= 115.0;         give it some negative values
      image[i] *= sqrt(2.0);     give it non-integer values! 
      }

  quantize_image (image, q_im, imsz, &bin_size, &range);

  printf("Range = %lf\n", range);

  for (pos=1,neg=1,i=0; (i<imsz) && (pos || neg); i++)
      { 
      if ( (pos) && (image[i] > 0.0) )
	  { printf("POS: "); print_binary(q_im[i]); printf("  "); pos=0; }
      if ( (neg) && (image[i] < 0.0) )
	  { printf("NEG: "); print_binary(q_im[i]); printf("  "); neg=0; }
      }
  printf("\n");

  embed_sign_bit(q_im, q_im, imsz);

  for (pos=1,neg=1,i=0; (i<imsz) && (pos || neg); i++)
      { 
      if ( (pos) && (image[i] > 0.0) )
	  { printf("POS: "); print_binary(q_im[i]); printf("  "); pos=0; }
      if ( (neg) && (image[i] < 0.0) )
	  { printf("NEG: "); print_binary(q_im[i]); printf("  "); neg=0; }
      }
  printf("\n");

  unembed_sign_bit(q_im, q_im, imsz);

  unquantize_image (q_im, res, imsz, bin_size);

  for (pos=1,neg=1,i=0; (i<imsz) && (pos || neg); i++)
      { 
      if ( (pos) && (image[i] > 0.0) )
	  { printf("POS: "); print_binary(res[i]); printf("  "); pos=0; }
      if ( (neg) && (image[i] < 0.0) )
	  { printf("NEG: "); print_binary(res[i]); printf("  "); neg=0; }
      }
  printf("\n");


  for (i=0; i<imsz; i++)
    image[i] = (float) res[i];

  printf("Writing result/data...\n");
  fp = check_fopen("result/data", "w");
  write_array(image, sizeof(*image), imsz, fp);  
  fclose(fp);

  check_free((char *) image);
  check_free((char *) q_im);
  check_free((char *) res);
  }

*/

/* Local Variables: */
/* compile-command: "gcc -o quantize  quantize.c utilities.c fileio.c -lm -lc" */
/* End: */

