/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <stdlib.h>
#include <math.h>
#include "epwic.h"
#include "run_length.h"
#include "prob_dist.h"
#include "arith.h"
#include "pyramid_util.h"

/* This extracts the abs value already sent to the child, the sign of the
   item, and the next bit to be sent (0 if not applicable) */
void get_sent_value(Coefficient value, int bit_level,
               int *abs, int*sign, int *next_bit){
  Coefficient mask;
  Coefficient sign_bit = 1<<(8*sizeof(Coefficient)-1);

  /* Get sign value */
  if (value & sign_bit){
    (*sign) = 1;
  } else {
    (*sign) = 0;
  }

  /* Get abs of value */
  value &= (~sign_bit);

  if (bit_level >= 0){
    mask = (1<<(bit_level+1))-1;
    *abs = value & (~mask);
    *next_bit = (value & (1<<bit_level))?1:0;
  } else {
    *abs = value;
    *next_bit = 0;
  }
}  


/* This returns the size of the encoded subband in bytes.  If the
   encoded size is bigger than the im_size/8, then we return -1. */

int arith_encode_subband(Pyramid pyr, int subband, int numbits, 
			 int *bit_to_send, int mult, arithcode * acode)
{
  int i;
  double parent_min,parent_max;
  int child_min,child_max;
  int child_bit_level;
  int sign,next_bit;
  int prob;
  int child_x, child_y;

  if (subband == 0) return -1; /* Useless on lowpass */

  child_bit_level =  bit_to_send[subband];

  for(i=0;i<pyrSubSize(pyr,subband);i++){
    /* get the values already sent, the sign, and the next bit */

    child_x = i % pyrSubXSize(pyr,subband);
    child_y = i / pyrSubXSize(pyr,subband);

    parent_min = 0;
    parent_max = MAX_ABS_COEFF+1;

    get_sent_value(pyrCoefCoord(pyr,subband,child_x,child_y),
                   child_bit_level,
		   &child_min,&sign,&next_bit);

    child_max = child_min + (1<<(child_bit_level+1));

    /* Calculate probability that the next bit is 0 */

    prob = (int) (1000 *
		  (get_prior_CDF(subband,child_min+(1<<child_bit_level))-
		   get_prior_CDF(subband,child_min))/
		  (get_prior_CDF(subband,child_max)-
		   get_prior_CDF(subband,child_min)));

    if (acode->lcd >= (acode->lcode - 2)){
      return (-1);
    } 

    /* If the division results in a probability we can't believe,
       set the probability to 1/2 */

    if ((prob <=0) || (prob >1000)) {
      prob = 500;
    }

    if (prob == 1000){
      prob = 999;
    }

    if (next_bit && child_min == 0){ /* encode sign bit as well */
      arith_encode(next_bit,prob,0,acode);
      arith_encode(sign,500,0,acode);
    } else {
      arith_encode(next_bit,prob,0,acode);
    }

  }

  return(acode->lcd);
}

/* This function calculates the contribution in terms of MSE of a given
   bitplane.  NOTE: THIS ASSUMES THE SCALE FACTOR IS 128 */

double bitplane_mse(Pyramid pyr, int subband, int bit_level){
  int i;
  double mse;
  double total_value, old_sent_value, new_sent_value;
  Coefficient sign_bit = 1<<(8*sizeof(Coefficient)-1);
  Coefficient value;

  mse = 0;
  for(i=0;i<pyr->sub_size[subband];i++){
    value = pyr->sub_coefs[subband][i];

    /* Get abs of value */
    value &= (~sign_bit);

    total_value = (double) value;
    old_sent_value = (double) ((value >> (bit_level+1)) << (bit_level+1));
    new_sent_value = (double) ((value >> (bit_level)) << (bit_level));
    total_value *= (double) pyr->sub_bin_size[subband] / 128.0;
    old_sent_value *= (double) pyr->sub_bin_size[subband] / 128.0;
    new_sent_value *= (double) pyr->sub_bin_size[subband] / 128.0;
    mse += (total_value - old_sent_value) * (total_value - old_sent_value) -
           (total_value - new_sent_value) * (total_value - new_sent_value);
  }
  return (mse);
}

/* get_subband_bit_level copies a specific bit level of a pyramid to
   dest_block.  The dest_block is a block of characters.  Please note
   that when we send the first significant bit, we immediately send the
   sign bit.  This helps in the creation of long zero runs */

int get_subband_bit_level(q_pyr, dest_block, pyr_offset, 
            im_size, bit_level)
     Coefficient *q_pyr;
     char* dest_block;
     int pyr_offset;
     int im_size;
     int bit_level;
{
  int i, out_bit;
  unsigned char symbol;
  int output_byte_loc;
  Coefficient sign_bit = 1<<(8*sizeof(Coefficient)-1);
  Coefficient value;
  int sign;

  output_byte_loc = 0;
  symbol = 0;
  out_bit = 7; /* We are copying into unsigned bits */
  for(i=0;i<im_size;i++){
    value = q_pyr[pyr_offset+i];

    /* Get sign value */
    if (value & sign_bit){
      sign = 1;
    } else {
      sign = 0;
    }

    /* Get abs of value */
    value &= (~sign_bit);

    if (value >> (bit_level+1)){  /* The significant bit
						   has been sent. */
      symbol |= ((value>>bit_level)&1)<<out_bit;
    } else {	/* We may be sending the most significant bit */
      if ((value>>bit_level)&1){ /* Send the sign bit too */
	symbol |= ((value>>bit_level)&1)<<out_bit;
	out_bit --;
	if (out_bit < 0){
               dest_block[(output_byte_loc)++] = symbol;
               symbol = 0;
               out_bit = 7;
             }
	symbol |= sign<<out_bit;
      } 
    }
    out_bit --;
    if (out_bit < 0){
      dest_block[(output_byte_loc)++] = symbol;
      symbol = 0;
      out_bit = 7;
    }
  }	
  if (out_bit != 7){
    dest_block[(output_byte_loc)++] = symbol;
  }
  return (output_byte_loc);
}

int estimate_encoding_size(Pyramid pyr, int subband, int *bit_to_send, 
			   arithcode *test_acode){

  int encoding_size,data_block_size,rle_block_size;
  unsigned char *rle_block;
  unsigned char *data_block;
  int pyr_offset;
  int numbits = 8 * sizeof(Coefficient);

  if (subband == 0){ /* Use rle encoding on lowpass */
    pyr_offset = 0;
  } else {
    pyr_offset = (((subband-1)%3)+1) * pyrSubSize(pyr,subband);
  }
  data_block = (unsigned char *) malloc(pyrSubSize(pyr,subband)*
					sizeof(Coefficient));
  data_block_size = get_subband_bit_level(pyr->sub_coefs[0],
					  data_block, pyr_offset,
					  pyrSubSize(pyr,subband),
					  bit_to_send[subband]);
  
  rle_block = (unsigned char *) malloc(data_block_size * 9 / 8);
  encoding_size = run_length_encode_zeros(data_block,
					  data_block_size,rle_block,1,2);
  if (data_block_size < encoding_size) encoding_size = data_block_size;
  if (subband > 0) {
    arcmak(256,test_acode);       /* Initialize acode */
    test_acode->codep = rle_block;  /* Set destination of encoding */
    test_acode->lcode = data_block_size * 9 / 8;
    rle_block_size = arith_encode_subband(pyr, subband, numbits, bit_to_send,
					  128,test_acode);
    if (rle_block_size > 0){
      if (rle_block_size < encoding_size) encoding_size = rle_block_size;
    }
  }
  free(data_block);
  free(rle_block);
  return (encoding_size);
}

/* reorders the bits of q_pyr (the quantized pyramid), run length
   encodes them using the best method available,  and places
   it into rq_pyr.  The size of rq_pyr must be at most the size of q_pyr +
   3*num_levels+1 bytes.... */
/* 10-2-96 We now try to minimize MSE by determining the subband whose
   next bitplane has the best change in MSE per coefficient sent */
/* 6-24-96 Stop if the worst error variance is better than 1/12 of the
   original sqared quantization level (which is one) */

int reorder_and_encode_pyr(Pyramid pyr, unsigned char *rq_pyr,int *salience,
			   int *bottom_bit, long outfile_size){
  int numbits,pyr_offset;
  int output_byte_loc;
  int data_block_size,rle_block_size;
  unsigned char *data_block;
  unsigned char *rle_block;
  int i;
  int min_block_size;
  int min_type;
  double *s,*n;  /* Used to calculate CDF */
  double *prior_s, *prior_p; /* Used to calculate Priors */
  int *sent_joint;  /* Have we sent joint info yet? */
  int *sent_prior; /* Have we sent prior info yet? */
  int *sent_bin_size; /* Have we sent the bin size yet? */
  int *bit_to_send;  /* What is the next bitplane to send? */
  double best_mse_ratio = 0; /* What is the best mse ratio calculated? */
  double *mse_ratio;     /* The last mse ratio for each subband */
  double worst_subband_error; /* What is the worst mse/coeff calculated? */
  double *subband_error; /* The mse/coeff calculated for each subband */
  double subband_mse;
  int best_mse_subband; /* What subband has the best mse ratio? */
  arithcode acode, test_acode;

  int center_mult, dif_mult, best_size,
      center_size, left_size, right_size;

  /* Initialize arith coder */
  acode.ilob=(unsigned long *)lvector(1,NWK);
  acode.iupb=(unsigned long *)lvector(1,NWK);
  arcmak(256,&acode);              /* Initialize acode */
  acode.codep = rq_pyr;           /* Set destination of encoding */
  acode.lcode = pyrImSize(pyr) * sizeof(unsigned long) + pyrNumSubs(pyr);

  /* Init part of test arith coder (used to see if arith coding is the best) */
  test_acode.ilob=(unsigned long *)lvector(1,NWK);
  test_acode.iupb=(unsigned long *)lvector(1,NWK);

  numbits = 8 * sizeof(Coefficient);
  output_byte_loc = 0;

  s = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));
  n = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));
  prior_s = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));
  prior_p = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));
  sent_joint = (int *) malloc(sizeof(int) * pyrNumSubs(pyr));
  sent_prior = (int *) malloc(sizeof(int) * pyrNumSubs(pyr));
  sent_bin_size = (int *) malloc(sizeof(int) * pyrNumSubs(pyr));
  bit_to_send = (int *) malloc(sizeof(int) * pyrNumSubs(pyr));
  mse_ratio = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));
  subband_error = (double *) malloc(sizeof(double) * pyrNumSubs(pyr));

  ForEachSubband(pyr,i){
    s[i] = 0;
    n[i] = 0;
    sent_joint[i] = 0;
    prior_s[i] = 0;
    prior_p[i] = 0;
    sent_prior[i] = 0;
    sent_bin_size[i] = 0;
    bit_to_send[i] = MAX_NUM_BITS - 1;
    mse_ratio[i] = -1;  /* Signify that it needs to be calculated */
    if (i>0){ /* Calculate probabilities for non-lowpass subbands */

      estimate_pdf(pyr, i, &prior_s[i], &prior_p[i]);
      prior_s[i] = floor((prior_s[i] - 0.5) / 128.0 * 255.0);
      prior_p[i] = floor((prior_p[i] - 0.2) / 1.8 * 255.0);
      if (prior_s[i] < 0)  prior_s[i] = 0.0;
      if (prior_s[i] > 255) prior_s[i] = 255.0;
      if (prior_p[i] < 0)   prior_p[i] = 0.0;
      if (prior_p[i] > 255) prior_p[i] = 255.0;
      calc_prior(i,prior_s[i] * 128.0 / 255.0 + 0.5,
		 prior_p[i] * 1.8 / 255.0 + 0.2);
    }
  }

  do {
    best_mse_subband = -1;
    worst_subband_error = -1;
    ForEachSubband(pyr,i){
      if (bit_to_send[i] >= bottom_bit[i]){

	if (i==0){
	  pyr_offset = 0;
	} else {
	  pyr_offset = (((i-1)%3)+1) * pyrSubSize(pyr,i);
	}

	if (mse_ratio[i] < 0){
	  subband_mse = bitplane_mse(pyr, i,bit_to_send[i]);
	  subband_error[i] = subband_mse / pyrSubSize(pyr,i);
	  mse_ratio[i] = subband_mse / 
	    estimate_encoding_size(pyr,i,bit_to_send,&test_acode);
	}
	if (worst_subband_error < 0){
	  worst_subband_error = subband_error[i];
	} else {
	  if (subband_error[i] > worst_subband_error){
	    worst_subband_error = subband_error[i];
	  }
	}
	if (best_mse_subband < 0){
	  best_mse_subband = i;
	  best_mse_ratio = mse_ratio[i];
	} else {
	  if (mse_ratio[i] > best_mse_ratio){
	    best_mse_subband = i;
	    best_mse_ratio = mse_ratio[i];
	  }
	}
      }
    }

    /* Stop if the worst error variance is better than 1/12 of the original
       sqared quantization level (which is one) */
    if (worst_subband_error <  (1.0/12.0)){
      break;
    }
    if (best_mse_subband >= 0){
      i = best_mse_subband;

      mse_ratio[i] = -1; /* mse_ratio needs to be recalculated */

      if (i==0){
	pyr_offset = 0;
      } else {
	pyr_offset = (((i-1)%3)+1) * pyrSubSize(pyr,i);
      }

      data_block = (unsigned char *) malloc(pyrSubSize(pyr,i)*
					    sizeof(Coefficient));
      data_block_size = get_subband_bit_level(pyr->sub_coefs[0], 
					      data_block, pyr_offset,
					      pyrSubSize(pyr,i), 
					      bit_to_send[i]);

      min_block_size = data_block_size;
      min_type = 0;

      rle_block = (unsigned char *) malloc(data_block_size * 9 / 8);
      rle_block_size = run_length_encode_zeros(data_block,
					       data_block_size,rle_block,1,2);
      if (min_block_size > rle_block_size){
	min_block_size = rle_block_size;
	min_type = 1;
      }

      rle_block_size = run_length_encode_zeros(data_block,
					       data_block_size,rle_block,2,2);
      if (min_block_size > rle_block_size){
	min_block_size = rle_block_size;
	min_type = 2;
      }

      arcmak(256,&test_acode);       /* Initialize acode */
      test_acode.codep = rle_block;  /* Set destination of encoding */
      test_acode.lcode = data_block_size * 9 / 8;
      center_mult = 128;
      dif_mult = 64;
      left_size = 0; right_size = 0;
      center_size = arith_encode_subband(pyr, i, numbits, bit_to_send, 
					    center_mult,&test_acode);
      best_size = center_size;
      rle_block_size = best_size;
      
      if (rle_block_size > 0){
	if (min_block_size > rle_block_size){
	  min_block_size = rle_block_size;
	  min_type = 3;
	}
      }
      
      if (min_type == 0){ /* raw */
	arith_encode_byte((unsigned char) (RAW_DATA_TAG | i), 0, &acode);
	if (!sent_bin_size[i]){
	  sent_bin_size[i] = 1;
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) >> 8),
			    0, &acode);
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) & 255),
			    0, &acode);
	}
	output_byte_loc = arith_encode_block(data_block, data_block_size, 
					     &acode);
      } else if (min_type == 1){ /* 8,16 */
	arith_encode_byte((unsigned char) (RLE_TYPE_ONE | i), 0, &acode);
	if (!sent_bin_size[i]){
	  sent_bin_size[i] = 1;
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) >> 8),
			    0, &acode);
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) & 255),
			    0, &acode);
	}
	rle_block_size = run_length_encode_zeros(data_block,
				    	 data_block_size,rle_block,1,2);
	output_byte_loc = arith_encode_block(rle_block, rle_block_size, 
					     &acode);
      } else if (min_type == 2){ /* 16,16 */
	arith_encode_byte((unsigned char) (RLE_TYPE_TWO | i), 0, &acode);
	if (!sent_bin_size[i]){
	  sent_bin_size[i] = 1;
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) >> 8),
			    0, &acode);
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) & 255),
			    0, &acode);
	}
	rle_block_size = run_length_encode_zeros(data_block,
					     data_block_size,rle_block,2,2);
	output_byte_loc = arith_encode_block(rle_block, rle_block_size, 
					     &acode);
      } else { /* arith encoding */
	arith_encode_byte((unsigned char) (RLE_TYPE_THREE | i), 0, &acode);
	if (!sent_bin_size[i]){
	  sent_bin_size[i] = 1;
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) >> 8),
			    0, &acode);
	  arith_encode_byte((unsigned char) (pyrSubBinSize(pyr,i) & 255),
			    0, &acode);
	}
	if (sent_prior[i] == 0){
	  sent_prior[i] = 1;
	  arith_encode_byte((unsigned char) (prior_s[i]), 0, &acode);
	  arith_encode_byte((unsigned char) (prior_p[i]), 0, &acode);
	}
	output_byte_loc = arith_encode_subband(pyr, i, numbits,bit_to_send,
					       center_mult,&acode);
      }
      free (rle_block);
      free (data_block);
      if (outfile_size > 0 && outfile_size < output_byte_loc){
	break;
      }
      bit_to_send[i] --; /* We sent a bitplane so decrease by 1 */
    }
  } while (best_mse_subband >=0);

  output_byte_loc = arith_encode_byte((unsigned char) DONE_DATA_TAG, 1,
				      &acode);

  free(s);
  free(n);
  free(prior_s);
  free(prior_p);
  free(sent_joint);
  free(sent_prior);
  free(sent_bin_size);
  free(bit_to_send);
  free(mse_ratio);
  free_lvector(test_acode.ilob,1,NWK);
  free_lvector(test_acode.iupb,1,NWK);
  free_lvector(acode.ilob,1,NWK);
  free_lvector(acode.iupb,1,NWK);

  return output_byte_loc;
}
