/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "epwic.h"
#include "arith.h"

/* This routine is used by run_length_encode_zeros() to read
   eight bits off of a string of unsigned characters so that
   they can be encoded */

unsigned char get_next_eight_bits(unsigned char *block, int block_size,
                        int *block_byte_loc, int *block_bit_loc){
  unsigned char val;
  int i;

  val = 0;
  for (i=7;i>=0 && ((*block_byte_loc)<block_size);i--){
    if (block[*block_byte_loc] & (1<<(7-(*block_bit_loc)))){
      val |= 1<<i;
    }
    (*block_bit_loc) ++;
    if ((*block_bit_loc) > 7){
      (*block_bit_loc) = 0;
      (*block_byte_loc)++;
    }
  }
  if ((*block_byte_loc)>=block_size){ /* return the rest of the bits as zero */
    (*block_bit_loc) += i+1;
    if ((*block_bit_loc) > 7){
       (*block_bit_loc) -=8;
       (*block_byte_loc)++;
    }
  }
  return (val);
}

/*
==========================================================================
run_length_encode_zeros() -- Right now in run length encodes
a block of chars into another block of chars.  Because
there is a possibility that run length encoding will expand a
string, the destination block 9/8ths the size of the initial block.
zero_block_size determines the size in bytes of the zero block
represented by each count in the zero block count.  raw_block_size
determines the size in bytes of the raw block.  For now, the size
used to hold the number of zero blocks is either one or two bytes
depending on the raw_block_size.
==========================================================================
*/

int run_length_encode_zeros (block, block_size, encoded_stream, 
                             zero_block_size, raw_block_size)
  register unsigned char *block;
  int block_size;
  register unsigned char *encoded_stream;
  int zero_block_size, raw_block_size;
  {
  unsigned char val;
  int block_byte_loc;
  int block_bit_loc;
  int output_byte_loc;
  int zero_run_mask = 1 << 7;
  int found_zero;
  unsigned long max_zeros;
  unsigned long num_zeros;
  int i;

  output_byte_loc = 0;
  block_byte_loc = 0;
  block_bit_loc = 0;

  if ((zero_block_size < 1) || (raw_block_size < 1)){
    fprintf(stderr,"run_length_encode_zeros: arguments out of range\n");
    return -1;
  }
  if (zero_block_size > 1){
    max_zeros = (1<<15)-1; /* Use two bytes */
  } else {
    max_zeros = (1<<7)-1;
  } 

  while(block_byte_loc < block_size){
    /* Try to find zeros */
    for(num_zeros = 0; num_zeros<max_zeros; num_zeros++){
      found_zero = 1;
      for(i=0;i<zero_block_size;i++){
	val = get_next_eight_bits(block, block_size, &block_byte_loc,
				  &block_bit_loc);
	found_zero &= (val == 0);
      }
      if (!found_zero){
	block_byte_loc -= zero_block_size; /* The block is not zeros so
                                              "unread" it */
        break;
      } else if (block_byte_loc >= block_size){
	num_zeros++;
	break;
      }
    }

    if (num_zeros > 0){  /* We found zeros!! */
      if (zero_block_size > 1){ /* Use two bytes to encode number */
	encoded_stream[output_byte_loc++] = zero_run_mask | 
	  (Byte) (num_zeros/256);
	encoded_stream[output_byte_loc++] = (Byte) (num_zeros%256);
      } else {           /* Use one byte to encode number */
	encoded_stream[output_byte_loc++] = zero_run_mask | (Byte) num_zeros;
      }
    } else {
      val = get_next_eight_bits(block, block_size, &block_byte_loc,
				&block_bit_loc);
      val>>=1;  /* In the first byte we can only save 7 bits */
      
      block_bit_loc --;  /* Move block location pointers back one bit */
      if (block_bit_loc < 0){ 
	block_bit_loc = 7;
	block_byte_loc --;
      }
      encoded_stream[output_byte_loc++] = val; /* record first byte */

      for (i=1;i<raw_block_size;i++){ /* loop to get the rest */
	val = get_next_eight_bits(block, block_size, &block_byte_loc,
				  &block_bit_loc);
	encoded_stream[output_byte_loc++] = val;
      }
    }
  }
  return (output_byte_loc);
}

/* The following are the RLE decode routines... Please note that they
   cannot be used on more then one RLE file at a time (actually it
   wouldn't take much to make them work on more than one file though) */

#define _RLE_RAW  0
#define _RLE_ENCODE  1

static FILE* _rle_infile = NULL;
static int _rle_raw_bit_buffer[15];
static unsigned long _rle_num_zeros;
static int _rle_raw_flag;
static unsigned int _rle_raw_in_buffer;
static unsigned int _rle_bytes_read;
static int _rle_type;
static int _rle_zero_block_size;
static int _rle_raw_block_size;
static int _bit_by_byte_loc;   /* Used to read bits a byte at a time */
static unsigned char _bit_by_byte_buffer;
static arithcode *_rle_acode;

/* If zero_block_size = 0, then we read the data raw.  Otherwise
   we follow the conventions for the encoding function above */

void rle_init_decode(FILE* infile, int zero_block_size, int raw_block_size,
		     arithcode *acode){
  _rle_raw_flag = 0;
  _rle_num_zeros = 0;
  _rle_raw_in_buffer = 0;
  _rle_bytes_read = 0;
  _rle_infile = infile;
  _rle_acode = acode;
  if (zero_block_size > 0){
	_rle_type = _RLE_ENCODE;
  } else {
	_rle_type = _RLE_RAW;
  }
  _rle_zero_block_size = zero_block_size;
  _rle_raw_block_size = raw_block_size;
  _bit_by_byte_loc = 0;
}

int rle_bytes_read(){
    return _rle_bytes_read;
}

/* return -1 on an error... */
int rle_fill_buffer(){
  unsigned char val;
  unsigned char zero_run_mask = 1 << 7;
  int i,j;

  _rle_bytes_read += arith_decode_file_byte(_rle_infile,&val,_rle_acode);

  if (_rle_type == _RLE_RAW){
		/* Put it in the buffer so that the highest number is
                   first */
    for (i=0;i<8;i++){
      _rle_raw_bit_buffer[i] = (val & (1 << i)) >> i;
    }
    _rle_raw_in_buffer = 8;
  } else {  
    if (val & zero_run_mask){              /* We have a run of zeros */
      _rle_raw_flag = 0;
      val &= ~zero_run_mask;
      if (_rle_zero_block_size > 1){ /* 16 bits for zero number block */
	_rle_num_zeros = val*256;
	_rle_bytes_read += arith_decode_file_byte(_rle_infile,&val,_rle_acode);
			/* Record the number of zeros */
        _rle_num_zeros = (_rle_num_zeros + val) * _rle_zero_block_size * 8;
      } else { /* 8 bits for zero number block */
	_rle_num_zeros = val * _rle_zero_block_size * 8;
      }
    } else {                      /* We have raw data, so read raw data */
      _rle_raw_flag = 1;
		/* Put it in the buffer so that the highest number is
                   first */
      for (i=0;i<7;i++){
	_rle_raw_bit_buffer[(_rle_raw_block_size-1)*8+i] = (val & (1<<i)) >> i;
      }
		/* Copy rest of raw block */
      for (j=2;j<=_rle_raw_block_size;j++){
	_rle_bytes_read += arith_decode_file_byte(_rle_infile,&val,_rle_acode);
	for (i=0;i<8;i++){
	  _rle_raw_bit_buffer[(_rle_raw_block_size-j)*8+i] = 
		(val & (1<<i)) >> i;
	}
      }
      _rle_raw_in_buffer = _rle_raw_block_size*8-1;
    }
  }
  return 0;  /* We succeeded */
}
			

/* On an error it returns -1 */
int rle_read_bit(int *bit){
  if (_rle_raw_flag || _rle_type == _RLE_RAW){	 /* we have raw data */
    if (_rle_raw_in_buffer == 0){ /* We're out of raw data */
      if (rle_fill_buffer() < 0){
	return -1;
      } else {
	return rle_read_bit(bit);
      }
    } else {
      (*bit) = _rle_raw_bit_buffer[_rle_raw_in_buffer-1];
      _rle_raw_in_buffer--;
    }
  } else { /* we have zero's */
    if (_rle_num_zeros == 0){  /* We're out of zeros */
      if (rle_fill_buffer() < 0){
	return -1;
      } else {
	return rle_read_bit(bit);
      }
    } else {
      (*bit) = 0;
      _rle_num_zeros--;
    }
  }
  return 0;
}

int rle_read_byte(unsigned char *sym){
  int i;
  int bit;
  unsigned char t_byte;
  
  t_byte = 0;
  for (i=7;i>=0;i--){  /* fill in byte left to right */
    if (rle_read_bit(&bit) < 0){
      (*sym) = 0;
      return -1;
    } else {
      t_byte <<= 1;
      t_byte |= (bit?1:0);
    }
  }
  (*sym) = t_byte;
  return 0;
}

/* Return only a bit, but read a whole byte in at a time */
int rle_read_bit_by_byte(int *bit){
  if (_bit_by_byte_loc == 0){
    if(rle_read_byte(&_bit_by_byte_buffer)<0){
      return -1;
    } else {
      _bit_by_byte_loc = 8;
    }
  }
  _bit_by_byte_loc --;
  (*bit) = (_bit_by_byte_buffer>>_bit_by_byte_loc & 1);
  return (*bit);
}

int rle_read_short(unsigned short *sym){
  unsigned char t_byte;
  unsigned short t_short;
  
  rle_read_byte(&t_byte); t_short = t_byte; t_short <<=8;
  rle_read_byte(&t_byte); t_short |= t_byte; (*sym)=t_short;
  return 0;
}

int rle_read_int(unsigned int *sym){
  unsigned char t_byte;
  unsigned int t_int;
  
  rle_read_byte(&t_byte); t_int = t_byte; t_int <<=8;
  rle_read_byte(&t_byte); t_int |= t_byte; t_int <<=8;
  rle_read_byte(&t_byte); t_int |= t_byte; t_int <<=8;
  rle_read_byte(&t_byte); t_int |= t_byte; (*sym)=t_int;
  return 0;
}

/* This will work with bytes, shorts and ints */
void rle_read_array(void *ptr, int ptrsize, int arraysize){
  int i;
  unsigned char *p;
  
  p = (unsigned char *) ptr;
  for (i=0;i<arraysize;i++,p = &(p[ptrsize])){
    if (ptrsize == 1){
      rle_read_byte(p);
    } else if (ptrsize == 2){
      rle_read_short((unsigned short *)p);
    } else if (ptrsize == 4){
      rle_read_int((unsigned int *) p);
    } else {
      fprintf(stderr,"rle_read_array:  Not 1, 2, or 4 bytes\n");
    }
  }
}

/*

void pr_ch(char c){
	int i;

	for (i=7;i>=0;i--){
		printf("%d",(c&(1<<i))>>i);
	}
}

void main(){
	FILE *infile;
	FILE *outfile;
	char string[] ="Thi\0\0\0\0s is a test\n";
	char new_string[256];
	int rle_length;
	int i;
	char t_byte;

	outfile = fopen ("test","w");
	rle_length = run_length_encode_zeros(string,16,new_string);
pr_ch(string[0]);
pr_ch(string[1]);
printf("\n");
pr_ch(new_string[0]);
pr_ch(new_string[1]);
printf("\n");
	fwrite(new_string,1,rle_length,outfile);
	fclose(outfile);

	infile = fopen("test","r");
	rle_init_decode(infile);
	for (i=0;i< 15;i++){
		rle_read_byte(&t_byte);
		printf("%d %d \n",string[i],t_byte);
	}
	fclose(outfile);
}

*/
	



















