/*
---------------------------------------------------------------------
---       EPWIC-1 (Embedded Predictive Wavelet Image Coder)       ---
---    Designed by Robert W. Buccigrossi and Eero P. Simoncelli   ---
---               Developed at the GRASP Laboratory               ---
---          Copyright 1997, University of Pennsylvania           ---
---                      All rights reserved.                     ---
---------------------------------------------------------------------

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice appear on all copies and
supporting documentation.  For any other uses of this software, in
original or modified form, including but not limited to distribution
in whole or in part, specific prior permission must be obtained from
U Penn and the authors.  These programs shall not be used, rewritten,
or adapted as the basis of a commercial software or hardware product
without first obtaining appropriate licenses from U Penn.  U Penn
makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied
warranty.

---------------------------------------------------------------------
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "collapse_pyr.h"
#include "quantize.h"
#include "parse_args.h"
#include "epwic.h"
#include "run_length.h"
#include "arith.h"
#include "pyramid_util.h"
#include "utilities.h"
#include "prob_dist.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_END 2
#endif

void save_image(outfile,q_pyr,x_size,y_size,num_levels,bin_size,scale_factor)
FILE *outfile;
Coefficient *q_pyr;
int x_size;
int y_size;
int num_levels;
BinValueType *bin_size;
float scale_factor;
{
    int *pyr;
    int *result;
    int i;
    int pixel_value;
    double dtemp;
    double min,new_scale_factor;

    /* -- Unquantize the pyramid -- */
    pyr = (int *) check_malloc( x_size*y_size*sizeof(*pyr) );
    unquantize_pyr(q_pyr, pyr, (x_size*y_size), num_levels, bin_size);

    /* -- Collapse pyramid, and divide by scale_factor -- */
    result = (int *) check_malloc (x_size*y_size*sizeof(*result));

    collapse_pyr(pyr, result, x_size, y_size, num_levels);
    min = 0.0;

    fprintf(outfile,"P5\n%d %d\n255\n",x_size,y_size);

    new_scale_factor = scale_factor;
    for (i=0; i<x_size*y_size; i++) {
        dtemp = (result[i] - min) / new_scale_factor;
        if (dtemp < 0) pixel_value = 0;              /* clip below */
        else if (dtemp > 255) pixel_value = 255;     /* clip above */
        else pixel_value = (int) (dtemp + 0.5);      /* round off  */
	putc(pixel_value,outfile);
    }

    fclose(outfile);
}

/* While we read in the data, we can be in two states:
        1) We need to read in the next type of data (or end of file)
        2) We are in the process of reading in the level of a pyramid
*/

#define GET_DATA_TYPE 1
#define READ_LEVEL    2
#define DONE          3

#define RLE_DECODE    1
#define ARITH_DECODE  2

int main(argc, argv)
  int argc;
  char *argv[];
  {
  FILE *infile,*outfile;
  Coefficient *q_pyr;
  Pyramid pyr;
  int x_size, y_size, num_levels;
  double scale_factor;
  Byte the_tag;
  int i;
  int next_symbol;
  int *read_to_bit;
  int subband,im_size = 0,im_offset,pyr_offset;
  Coefficient est_mask;
  int state;
  int bytes_read;
  int old_bytes_read;
  int bin_bytes_read;
  unsigned long bytes_to_read;
  int message_time;
  int read_sign;

  arithcode acode;
  int decode_type = 0;
  unsigned char temp_char;

  scale_factor = 128.0;  /* Hard coded right now */

  /* --- Parse arglist, Reading the data from the input file, unpacking it, 
     and decoding it, and putting it into the quantized pyramid --- */
  parse_unepwic_args(argc, argv, &infile, &num_levels, &x_size, &y_size,
            &outfile, &bytes_to_read);

  /* Initialize pyramid datastructure */

  q_pyr = (Coefficient *) check_malloc( x_size*y_size*sizeof(*q_pyr) );
  bzero((char*) q_pyr, x_size * y_size * sizeof(Coefficient));
  pyr = pyrNewPyramid(q_pyr, x_size, y_size, num_levels);

  init_prior(num_levels);

  /* Init arithmetic coding */
  acode.ilob=(unsigned long *)lvector(1,NWK);
  acode.iupb=(unsigned long *)lvector(1,NWK);
  acode.codep = (unsigned char *)malloc(NWK * sizeof(unsigned char));
  acode.lcode = NWK;
  arcmak(256,&acode);

  /* BIN INFO NOW READ AT LAST MOMENT
     read_byte(the_tag, infile);
     if (the_tag != (Byte)BIN_INFO_TAG) {
       fprintf(stderr,"Expecting BIN_INFO first...\n");
       exit(-1);
     }
     ForEachSubband(pyr,i){
       read_short(pyrSubBinSize(pyr,i), infile);
     }
  */

  ForEachSubband(pyr,i){
    pyrSubBinSize(pyr,i) = 0;
  }

  /* -- now to start the whole mess -- */

  read_to_bit = (int *) check_malloc((3*num_levels+1)*sizeof(*read_to_bit));
  for(i=0;i<3*num_levels+1;i++){
	read_to_bit[i] = (int) ceil(log2((double)MAX_ABS_COEFF))-1;
  }

  state = GET_DATA_TYPE;
  subband = 0;     /* Level of the pyrmid we are reading */
  pyr_offset = 0; /* Location of the level in the array */
  im_offset = 0;  /* Location in the level we are at */
  bytes_read = 9; /* Calculate size of header */
  old_bytes_read = bytes_read;
  bin_bytes_read = 0;
  message_time = 256;
  read_sign = 0;

  while(state != DONE){

    if ((bytes_to_read > 0) && (bytes_read >= bytes_to_read)){
      save_image(outfile,q_pyr,x_size,y_size,num_levels,
		 pyr->sub_bin_size,scale_factor);
      printf("Read to %d bytes\n",(int)ftell(infile));
      exit(0);
    }
    
    if (state == GET_DATA_TYPE){

      bytes_read += arith_decode_file_byte(infile,&the_tag,&acode);
      old_bytes_read = bytes_read;
      bin_bytes_read = 0;

      if (the_tag == (Byte) DONE_DATA_TAG){ 
	state = DONE;
      } else if (the_tag & (Byte) DATA_MASK){

	state = READ_LEVEL;
        subband = the_tag & 0x1F;

	/* Get Bin Info if needed */
	if(pyrSubBinSize(pyr,subband) == 0){
	  bin_bytes_read += arith_decode_file_byte(infile,&temp_char,&acode);
	  pyrSubBinSize(pyr,subband) = ((unsigned short)temp_char) << 8;
	  bin_bytes_read += arith_decode_file_byte(infile,&temp_char,&acode);
	  pyrSubBinSize(pyr,subband) += ((unsigned short)temp_char);
	}

	im_offset = 0;
        if (subband==0){
	  im_size = (x_size * y_size) >> (2*num_levels);
	  pyr_offset = 0;
        } else {
	  im_size = (x_size * y_size) >> (2*(num_levels-((subband-1)/3)));
	  pyr_offset = (((subband-1)%3)+1) * im_size;
        }

	if ((the_tag & (Byte) DATA_MASK) == (Byte) RAW_DATA_TAG){
	  rle_init_decode(infile,0,0,&acode); /* raw */
	  decode_type = RLE_DECODE;
	} else if ((the_tag & (Byte) DATA_MASK) == (Byte) RLE_TYPE_ONE){
	  rle_init_decode(infile,1,2,&acode); /* 8,16 */
	  decode_type = RLE_DECODE;
	} else if ((the_tag & (Byte) DATA_MASK) == (Byte) RLE_TYPE_TWO){
	  rle_init_decode(infile,2,2,&acode); /* 16,16 */
	  decode_type = RLE_DECODE;
	} else { /* arith decode*/
	  arith_bitplane_init_decode(infile,pyr,subband,
				     read_to_bit,&acode); 
	  decode_type = ARITH_DECODE;
	}

	/* Initiate estimation lookup table */
	init_estimation_fraction(subband,read_to_bit[subband]);

      } else {
        fprintf(stderr,"Unknown data tag... ignoring\n");
      } 
    } else if (state == READ_LEVEL){

      if (decode_type == RLE_DECODE){
	rle_read_bit_by_byte(&next_symbol);
	bytes_read = old_bytes_read + rle_bytes_read() + bin_bytes_read;
      } else {
	next_symbol = arith_bitplane_read_bit();
	bytes_read = old_bytes_read + arith_bitplane_bytes_read() +
                     bin_bytes_read;
      }
      
      if (read_sign){	  /* Store the sign bit */

	read_sign = 0;
	pyrCoefOffset(pyr,subband,im_offset) |=
	    ((next_symbol)&1)*COEF_SIGN_BIT;

	/* Here we make an estimate about the rest of the number */
	if ((read_to_bit[subband] > 0) && 
	    (pyrCoefOffset(pyr,subband,im_offset) & ~((Coefficient) 
				   ((1<<(read_to_bit[subband])) -1)))){

	  if (read_to_bit[subband] < MAX_NUM_BITS){
	    /* Remove old estimation */
	    pyrCoefOffset(pyr,subband,im_offset) &=
	      ~((Coefficient) (1<<(read_to_bit[subband])) - 1);
	  }
	  est_mask = get_estimation_fraction(
		     (pyrCoefOffset(pyr,subband,im_offset) & ~(COEF_SIGN_BIT))
		     >> read_to_bit[subband]);
				   
	  pyrCoefOffset(pyr,subband,im_offset) |= est_mask;
	}
	im_offset ++;  /* With the sign read, we can move on */

      } else if (pyrCoefOffset(pyr,subband,im_offset) >> 
		 (read_to_bit[subband]+1)){
	/* significant bit sent */

	/* First remove the value at the bit to which we are
	   storing the new value */
	pyrCoefOffset(pyr,subband,im_offset) &=
	              ~((Coefficient) 1<<(read_to_bit[subband]));

	/* Store the new bit read */
	pyrCoefOffset(pyr,subband,im_offset) |= 
	    ((next_symbol)&1)<<(read_to_bit[subband]);

	/* Here we make an estimate about the rest of the number */
	if ((read_to_bit[subband] > 0) && 
	    (pyrCoefOffset(pyr,subband,im_offset) & ~((Coefficient) 
				   ((1<<(read_to_bit[subband])) -1)))){

	  if (read_to_bit[subband] < MAX_NUM_BITS){
	    /* Remove old estimation */
	    pyrCoefOffset(pyr,subband,im_offset) &=
	      ~((Coefficient) (1<<(read_to_bit[subband])) - 1);
	  }
	  est_mask = get_estimation_fraction(
		     (pyrCoefOffset(pyr,subband,im_offset) & ~(COEF_SIGN_BIT))
		     >> read_to_bit[subband]);

	  pyrCoefOffset(pyr,subband,im_offset) |= est_mask;
	}
	im_offset ++;
      } else { /* significant bit not sent */
	if ((next_symbol)&1) { /* reached significant bit */
	  pyrCoefOffset(pyr,subband,im_offset) |= 
	    ((next_symbol)&1)<<(read_to_bit[subband]);
	  read_sign = 1; /* Read sign bit next */
	  /* Do NOT increment im_offset since we need to read sign bit */
	} else { /* not reached significant bit */
	  im_offset ++;
	}
      }
      if (im_offset >= im_size){
	state = GET_DATA_TYPE;
	read_to_bit[subband]--;
      }
    }
  }

  save_image(outfile,q_pyr,x_size,y_size,num_levels,
	     pyr->sub_bin_size,scale_factor);
  
  fclose(infile);
  
  check_free ((char *) q_pyr);
  
  return(1);
}


