/*
 * ========================================================================
 * xfb.c - library of routines for non-X programs that want to do 
 *         quick X graphics, without dramatically re-writing the code
 *         currently requires an 8bit color X display
 *
 *  Author:    John Bradley (bradley@cis.upenn.edu)
 *
 *  Modified:  Hany Farid
 *             Robert Buccigrossi - added more intelligent grayscale
 *                                  allocation
 * ========================================================================
 */

#include <stdlib.h>
#include "xfb_byte.h"

/* X stuff */
static int           theScreen;
static unsigned int  ncells, dispWIDE, dispHIGH, dispDEEP;
static Colormap      theCmap;
static Display       *display;
static Window        rootW;
static GC            theGC;
static Visual        *theVisual;
unsigned long        white, black;
XFontStruct          *mfinfo;
Font                 mfont;

/* not directly X stuff */
static char *wname;
static Window createWindow();

/* color table stuff */
unsigned long grays[NUMGRAY];
short gray_allocated[NUMGRAY];


void GetGreyscale(unsigned long *grays,int Depth)
     /***************************************
      * Get the requested greyscale
      */
{
  XColor    xc;
  int       i,j,k;
  int       failed;
  int       num_allocated = 2;  /* Black and White */
 
  /*
   * First Pass
   */
  
  /* Specify the first color: black */
  xc.red = xc.green = xc.blue = 0;
  
  /* Allocate the color, it is already in the colormap */
  XAllocColor(display,theCmap,&xc);
  
  grays[0] = xc.pixel;
  gray_allocated[0] = 1;

  /* Do the same for white */
  xc.red = xc.green = xc.blue = 65535;
  XAllocColor(display,theCmap,&xc);
  grays[NUMGRAY-1] = xc.pixel;
  gray_allocated[NUMGRAY-1] = 1;
  
  /* Set the rest of the colors as the same */
  for (i=1;i<NUMGRAY-1;i++){
    grays[i] = grays[0];
    gray_allocated[i] = 0;
  }
  
  /*
   * Final Passes, get as many greyscales as possible
   */

  failed = 0;

  if (Depth > 1){
    for (k=NUMGRAY;(k>1) && (!failed);k/=2)
      for (i=k/2;(i<NUMGRAY) && (!failed);i+=k) {
	
	xc.red = xc.green = xc.blue = (i*0xffff)/256;
	if (!XAllocColor(display,theCmap,&xc)){
	  failed = 1;
	  continue;
	}
	
	grays[i] = xc.pixel;
	gray_allocated[i] = 1;
        num_allocated ++;
      }
      if (num_allocated < 16){
	fprintf(stderr,"Only able to allocate %d greys\n",num_allocated);
      }
  } else {   /* Monocrome */
    grays[0] = 0;
    grays[NUMGRAY] = 255;
  }

  
  /* Now, if we haven't allocated all the colors, we point the
     colors not allocated to colors that are */
  j=0;
  for (i=0;i<NUMGRAY;i++){ /* Search forwards */
    if (gray_allocated[i]){
      j=i;
    } else {
      grays[i] = j;
    }
  }
  for (i=NUMGRAY-1;i>=0;i--){ /* Now search backwards */
    if (gray_allocated[i]){
      j=i;
    } else {
      if ((j-i) < (i-grays[i])){  /* The distance is smaller */
	grays[i] = j;
      }
    }
  }

  /* done */
  return;
}

void DitherOneD(byte *epic, int height, int bperlin,int pad){
  int i, j, error_val;
  int new_val;

  error_val = 0;
  
  for (i=0; i<height;i++){
    error_val = 0;

    for(j=0;j<bperlin-pad;j++){
      new_val = epic[i*bperlin + j] + error_val;
      error_val = 0;
      
      if (new_val >= NUMGRAY){ /* new_val is too high */
	error_val = new_val - NUMGRAY - 1;
	new_val = NUMGRAY - 1;
      } else if (new_val < 0){ /* new_val is too low */
	error_val = new_val;
	new_val = 0;
      }
      
      if (gray_allocated[new_val]){
	epic[i*bperlin + j] = grays[new_val];
      } else { /* grays[new_val] points to closest color; */
	error_val += new_val - grays[new_val];
	epic[i*bperlin + j] = grays[grays[new_val]];
      }
    }
  }
}

void DitherTwoD(byte *epic, int height, int bperlin,int pad){
  int i, j, error_val;
  int new_val;
  int *vert_error_val;

  vert_error_val = (int *) malloc(sizeof(int) * bperlin);

  for (i=0;i<bperlin;i++){
    vert_error_val[i] = 0;
  }

  for (i=0; i<height;i++){
    error_val = 0;
    for(j=0;j<bperlin-pad;j++){
/*
      new_val = epic[i*bperlin + j] + (error_val/4) + (vert_error_val[j]/4);
*/
      new_val = epic[i*bperlin + j] + (vert_error_val[j]/4);
      if (j>0) new_val += (vert_error_val[j-1]/4);
      if (j<bperlin-1) new_val +=(vert_error_val[j+1]/4);
      error_val = 0;
      vert_error_val[j] = 0;
      
      if (new_val >= NUMGRAY){ /* new_val is too high */
	error_val = new_val - NUMGRAY - 1;
	new_val = NUMGRAY - 1;
      } else if (new_val < 0){ /* new_val is too low */
	error_val = new_val;
	new_val = 0;
      }
      
      if (gray_allocated[new_val]){
	epic[i*bperlin + j] = grays[new_val];
      } else { /* grays[new_val] points to closest color; */
	error_val += new_val - grays[new_val];
	epic[i*bperlin + j] = grays[grays[new_val]];
      }

      error_val = error_val;
      vert_error_val[j] = error_val;
    }
  }

  free(vert_error_val);
}

void ConvertToMonochrome(byte* epic,int height,int bperlin,int pad){
  int i,j,dest,dest_bit;
  int dest_height;

  dest_bit = 1<<7;
  dest = 0;
  dest_height = 0;
  for (i=0; i<height;i++){
    for(j=0;j<bperlin-pad;j++){
      if (epic[i*bperlin + j]){
	epic[dest] = epic[dest] | dest_bit;
      } else {
	epic[dest] = epic[dest] & (~dest_bit);
      }
      if (dest_bit == 1){
	dest_bit = 1<<7;
	dest++;
	if ((((dest*8) % bperlin) > (bperlin - pad)) ||
	    (((dest*8) % bperlin) == 0)){
	  dest_height ++;
	  dest = dest_height * bperlin;
	}
      } else {
	dest_bit = dest_bit>>1;
      }
    }
  }
}

/*
 * ========================================================================
 * InitX(): initialize an X window.
 * ========================================================================
 */
void InitX( int idispwin, char *name, int xpos, int ypos,int w,int h )
{
	Window window;

	wname = name;

	if( (display = XOpenDisplay( NULL )) == NULL ) {
		fprintf( stderr, "Can't open display\n" );
		exit( -1 );
	}

	theScreen = DefaultScreen(display);
	theCmap   = DefaultColormap(display, theScreen);
	rootW     = RootWindow(display,theScreen);
	theGC     = DefaultGC(display,theScreen);
	theVisual = DefaultVisual(display,theScreen);
	ncells    = DisplayCells(display, theScreen);
	dispWIDE  = DisplayWidth(display,theScreen);
	dispHIGH  = DisplayHeight(display,theScreen);
	dispDEEP  = DisplayPlanes(display,theScreen);

	white = WhitePixel(display,theScreen);
	black = BlackPixel(display,theScreen);

	window = createWindow( display, xpos, ypos, w, h );
	XMapWindow( display, window );

	/* Generate a colormap */
	GetGreyscale(grays,dispDEEP);

	/* Load font */
	if( (mfinfo = XLoadQueryFont( display, FONT1 ) ) != NULL ) {
	  mfont = mfinfo->fid;
	  XSetFont( display, theGC, mfont );
	}

	XFlush( display );

	dispwin[ idispwin ].display = display;
	dispwin[ idispwin ].window = window;
}

/*
 * ========================================================================
 * createWindow():  called by InitX().
 * ========================================================================
 */
static Window createWindow( Display * display, int xpos,int ypos,int w,int h )
{
	Window window;
	XSetWindowAttributes xswa;
	unsigned int         xswamask;
	XSizeHints           hints;
	XTextProperty		 text_prop;

	hints.x = xpos;        hints.y = ypos; 
	hints.width = w;       hints.height = h;
	hints.max_width  = w;  hints.max_height = h;
	hints.min_width  = w;  hints.min_height = h;
	hints.flags = USPosition | USSize | PMaxSize | PMinSize;

	xswa.background_pixel = black;
	xswa.border_pixel     = white;
	xswamask = CWBackPixel | CWBorderPixel;

	window = XCreateWindow( display, rootW, hints.x, hints.y, w, h, 2, 0,
	  CopyFromParent, CopyFromParent, xswamask, &xswa );
	if( !window ) {
		fprintf( stderr, "Can't create window" );
		exit(-1);
	}

    xswa.event_mask = ExposureMask | ButtonPressMask;
/*
    attr.event_mask = ExposureMask | ButtonReleaseMask | ButtonPressMask |
      KeyPressMask | KeyReleaseMask | StructureNotifyMask;
*/
    XChangeWindowAttributes(display, window, CWEventMask, &xswa);

	XStringListToTextProperty(&wname,1,&text_prop);
	XSetWMProperties(display, window, &text_prop, &text_prop, NULL, 0, 
			&hints, NULL, NULL);

	return( window );
}

/*
 * ========================================================================
 * DrawImage8(): blast an 8-bit image to an X window.
 * ========================================================================
 */
static XImage *ximage = NULL;
static byte   *epic   = NULL;
static long   lastsize = 0;

void DrawImage8( int idispwin, byte *p, int x, int y, int w, int h )
{
	int i, bperlin, pad;
	byte *ep;
	long size;
	Display *display;
	Window window;

	display = dispwin[ idispwin ].display;
	window = dispwin[ idispwin ].window;

	/* note: have to pad epic so that it's a multiple of 32-bits per line */
	size = w * h;
	bperlin = (w + 3) & (~3);
	pad = bperlin - w;

	if( size != lastsize || !epic ) { /* have to re-allocate epic */
		if( epic )
			free( epic );
		epic = (byte *)malloc( bperlin * h );
		if( !epic ) { 
			fprintf( stderr, "Can't malloc epic\n" ); 
			exit( -1 ); 
		}
	}

	if( !pad ) {
		memcpy(epic, p, size);
	} else {
		for (i=0; i<h; i++) {
			ep = epic + (i*bperlin);
			memcpy(ep, p, bperlin-pad);
			p += w;
		}
	}

	/* We have to fix dealing with the pad */
	DitherTwoD(epic,h,bperlin,pad);

	if (dispDEEP == 1) ConvertToMonochrome(epic,h,bperlin,pad);
	
	if (!ximage || size != lastsize) {    /* create XImage  */
	  if( ximage ) 
	    XDestroyImage(ximage);
	  ximage = XCreateImage( display, theVisual, dispDEEP, ZPixmap, 0,
				(char *)epic, w, h, 32, bperlin );
	}

	XPutImage( display, window, theGC, ximage, 0, 0, x, y, w, h );
	lastsize = size;
}

/* Note: doesn't work since we changed the gray allocation */
void DrawLine(int idispwin, int x,int  y,int x1,int y1,int c)
{
  Display *display;
  Window window;

  display = dispwin[ idispwin ].display;
  window = dispwin[ idispwin ].window;

  c >>= GRAYSHIFT;   /* shift down to # of grays we have */
  /* should actually set color to cols[c] */
  XSetForeground(display,theGC, grays[c]);

  XDrawLine(display, window, theGC, x, y, x1, y1);
  XFlush(display);
}

int ButtonPressed(int idispwin)
{
  Display *display;
  Window window;
  XEvent event;

  display = dispwin[ idispwin ].display;
  window = dispwin[ idispwin ].window;

  if (XCheckTypedEvent(display,ButtonPress,&event)){
	return(event.xbutton.button);
  } else {
    return 0;
  }
}

int WindowExposure(int idispwin)
{
  Display *display;
  Window window;
  XEvent event;

  display = dispwin[ idispwin ].display;
  window = dispwin[ idispwin ].window;

  if (XCheckTypedEvent(display,Expose,&event)){
	while(XCheckTypedEvent(display,Expose,&event)); 
                                        /* Remove all Expose events */
	return 1;
  } else {
    return 0;
  }
}

void WriteText(int idispwin,char* text,int x, int y)
{
  Display *display;
  Window window;

  display = dispwin[ idispwin ].display;
  window = dispwin[ idispwin ].window;

  XSetForeground(display,theGC, grays[255]);
 
  XDrawString(display,window,theGC,x,y,text,strlen(text));
} 

void DrawBlackBox(int idispwin, int x, int y, int w, int h)
{
  Display *display;
  Window window;
  int i;

  display = dispwin[ idispwin ].display;
  window = dispwin[ idispwin ].window;

  XSetForeground(display,theGC, grays[0]);

  for (i=0;i<h;i++){
    XDrawLine(display,window,theGC,x,y+i,x+w,y+i);
  }
}
